"""
Test GenericProvider
"""

import unittest

from mock import MagicMock, patch

from sickchill.providers.GenericProvider import GenericProvider


class GenericProviderTests(unittest.TestCase):
    """
    Test GenericProvider
    """

    def test_get_id(self):
        """
        Test get_id
        """
        test_cases = {
            None: "",
            123: "123",
            12.3: "12_3",
            0: "",
            -123: "_123",
            -12.3: "_12_3",
            "": "",
            " ": "",
            "123": "123",
            " 123 ": "123",
            "12.3": "12_3",
            " 12.3 ": "12_3",
            "0": "0",
            " 0 ": "0",
            "-123": "_123",
            " -123 ": "_123",
            "-12.3": "_12_3",
            " -12.3 ": "_12_3",
            "abc": "abc",
            " abc ": "abc",
            "ABC": "abc",
            " ABC ": "abc",
            ".def": "_def",
            "g,hi": "g_hi",
            "jk!l": "jk_l",
            "mno?": "mno_",
            "_pqr$": "_pqr_",
        }

        unicode_test_cases = {
            "": "",
            " ": "",
            "123": "123",
            " 123 ": "123",
            "12.3": "12_3",
            " 12.3 ": "12_3",
            "0": "0",
            " 0 ": "0",
            "-123": "_123",
            " -123 ": "_123",
            "-12.3": "_12_3",
            " -12.3 ": "_12_3",
            "abc": "abc",
            " abc ": "abc",
            "ABC": "abc",
            " ABC ": "abc",
            ".def": "_def",
            "g,hi": "g_hi",
            "jk!l": "jk_l",
            "mno?": "mno_",
            "_pqr$": "_pqr_",
        }

        for test in test_cases, unicode_test_cases:
            for (name, result) in test.items():
                assert GenericProvider(name).get_id() == result

    def test_image_name(self):
        """
        Test image_name
        """
        test_cases = {
            None: ".png",
            123: "123.png",
            12.3: "12_3.png",
            0: ".png",
            -123: "_123.png",
            -12.3: "_12_3.png",
            "": ".png",
            " ": ".png",
            "123": "123.png",
            " 123 ": "123.png",
            "12.3": "12_3.png",
            " 12.3 ": "12_3.png",
            "0": "0.png",
            " 0 ": "0.png",
            "-123": "_123.png",
            " -123 ": "_123.png",
            "-12.3": "_12_3.png",
            " -12.3 ": "_12_3.png",
            "abc": "abc.png",
            " abc ": "abc.png",
            "ABC": "abc.png",
            " ABC ": "abc.png",
            ".def": "_def.png",
            "g,hi": "g_hi.png",
            "jk!l": "jk_l.png",
            "mno?": "mno_.png",
            "_pqr$": "_pqr_.png",
        }

        unicode_test_cases = {
            "": ".png",
            " ": ".png",
            "123": "123.png",
            " 123 ": "123.png",
            "12.3": "12_3.png",
            " 12.3 ": "12_3.png",
            "0": "0.png",
            " 0 ": "0.png",
            "-123": "_123.png",
            " -123 ": "_123.png",
            "-12.3": "_12_3.png",
            " -12.3 ": "_12_3.png",
            "abc": "abc.png",
            " abc ": "abc.png",
            "ABC": "abc.png",
            " ABC ": "abc.png",
            ".def": "_def.png",
            "g,hi": "g_hi.png",
            "jk!l": "jk_l.png",
            "mno?": "mno_.png",
            "_pqr$": "_pqr_.png",
        }

        for test in test_cases, unicode_test_cases:
            for (name, result) in test.items():
                assert GenericProvider(name).image_name() == result

    def test_is_active(self):
        """
        Test is_active
        """
        assert not GenericProvider("Test Provider").is_active

    def test_is_enabled(self):
        """
        Test is_enabled
        """
        assert not GenericProvider("Test Provider").is_enabled

    def test_make_id(self):
        """
        Test make_id
        """
        test_cases = {
            None: "",
            123: "123",
            12.3: "12_3",
            0: "",
            -123: "_123",
            -12.3: "_12_3",
            "": "",
            " ": "",
            "123": "123",
            " 123 ": "123",
            "12.3": "12_3",
            " 12.3 ": "12_3",
            "0": "0",
            " 0 ": "0",
            "-123": "_123",
            " -123 ": "_123",
            "-12.3": "_12_3",
            " -12.3 ": "_12_3",
            "abc": "abc",
            " abc ": "abc",
            "ABC": "abc",
            " ABC ": "abc",
            ".def": "_def",
            "g,hi": "g_hi",
            "jk!l": "jk_l",
            "mno?": "mno_",
            "_pqr$": "_pqr_",
        }

        unicode_test_cases = {
            "": "",
            " ": "",
            "123": "123",
            " 123 ": "123",
            "12.3": "12_3",
            " 12.3 ": "12_3",
            "0": "0",
            " 0 ": "0",
            "-123": "_123",
            " -123 ": "_123",
            "-12.3": "_12_3",
            " -12.3 ": "_12_3",
            "abc": "abc",
            " abc ": "abc",
            "ABC": "abc",
            " ABC ": "abc",
            ".def": "_def",
            "g,hi": "g_hi",
            "jk!l": "jk_l",
            "mno?": "mno_",
            "_pqr$": "_pqr_",
        }

        for test in test_cases, unicode_test_cases:
            for (name, result) in test.items():
                assert GenericProvider.make_id(name) == result

    def test_seed_ratio(self):
        """
        Test seed_ratio
        """
        assert GenericProvider("Test Provider").seed_ratio() == ""

    def test__check_auth(self):
        """
        Test _check_auth
        """
        assert GenericProvider("Test Provider")._check_auth()

    def test_login(self):
        """
        Test login
        """
        assert GenericProvider("Test Provider").login()

    def test_search(self):
        """
        Test search
        """
        test_cases = {
            None: [],
            123: [],
            12.3: [],
            -123: [],
            -12.3: [],
            "": [],
            "123": [],
            "12.3": [],
            "-123": [],
            "-12.3": [],
        }

        unicode_test_cases = {
            "": [],
            "123": [],
            "12.3": [],
            "-123": [],
            "-12.3": [],
        }

        for test in test_cases, unicode_test_cases:
            for (search_params, result) in test.items():
                assert GenericProvider("Test Provider").search(search_params) == result

    def test__get_size(self):
        """
        Test _get_size
        """
        assert GenericProvider("Test Provider")._get_size(None) == -1

    def test__get_storage_dir(self):
        """
        Test _get_storage_dir
        """
        assert GenericProvider("Test Provider")._get_storage_dir() == ""

    def test__get_title_and_url(self):
        """
        Test _get_title_and_url
        """
        items_list = [
            None,
            {},
            {"link": None, "title": None},
            {"link": "", "title": ""},
            {"link": "http://www.google.com/&amp;foo=bar%26tr%3Dtest", "title": "Some Title"},
        ]
        results_list = [("", ""), ("", ""), ("", ""), ("", ""), ("Some.Title", "http://www.google.com/&foo=bar&tr=test")]

        unicode_items_list = [{"link": "", "title": ""}, {"link": "http://www.google.com/&amp;foo=bar%26tr%3Dtest", "title": "Some Title"}]
        unicode_results_list = [("", ""), ("Some.Title", "http://www.google.com/&foo=bar&tr=test")]

        assert len(items_list) == len(results_list), "Number of parameters ({0:d}) and results ({1:d}) does not match".format(
            len(items_list), len(results_list))

        assert len(unicode_items_list) == len(unicode_results_list), "Number of parameters ({0:d}) and results ({1:d}) does not match".format(
            len(unicode_items_list), len(unicode_results_list))

        for (index, item) in enumerate(items_list):
            assert GenericProvider("Test Provider")._get_title_and_url(item) == results_list[index]

        for (index, item) in enumerate(unicode_items_list):
            assert GenericProvider("Test Provider")._get_title_and_url(item) == unicode_results_list[index]

    def test__verify_download(self):
        """
        Test _verify_download
        """
        assert GenericProvider("Test Provider")._verify_download("Random.torrent")

    @patch("sickchill.providers.GenericProvider.download_file")
    @patch("sickchill.providers.GenericProvider.remove_file_failed")
    def test_download_file(self, remove_file_mock, df_mock):
        """
        Test download_result
        """
        domain = "domain"
        filename = "TestFilename.nzb"
        urls = [
            "http://{0}/{1}.torrentNO_DOWNLOAD_NAME".format(domain, filename),
            "http://{0}/{1}.torrent".format(domain, filename),
        ]

        # Test the login() check
        gp1 = GenericProvider("Test Provider 1")
        login_mock = MagicMock()
        login_mock.return_value = False
        with patch.object(gp1, "login", login_mock):
            assert not gp1.download_result("result 1")
            assert login_mock.called

        # Test the _make_url call
        gp2 = GenericProvider("Test Provider 2")
        make_url_mock = MagicMock()
        make_url_mock.return_value = (urls, filename)
        df_mock.return_value = True
        with patch.object(gp2, "_make_url", make_url_mock):
            resp = gp2.download_result("result 2")
            assert resp
            assert "Referer" in gp2.headers
            assert domain in gp2.headers["Referer"]
            assert df_mock.called

        # Test the remove_file_failed path
        gp3 = GenericProvider("Test Provider 3")
        make_url_mock = MagicMock()
        make_url_mock.return_value = (urls, filename)
        verify_download_mock = MagicMock()
        verify_download_mock.return_value = False
        df_mock.return_value = True
        with patch.object(gp3, "_make_url", make_url_mock):
            with patch.object(gp3, "_verify_download", verify_download_mock):
                resp = gp3.download_result("result 3")
                assert not resp
                assert remove_file_mock.called


if __name__ == "__main__":
    print("=====> Testing {0}".format(__file__))

    SUITE = unittest.TestLoader().loadTestsFromTestCase(GenericProviderTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



"""
Test history
"""

import unittest

from sickchill.oldbeard.common import Quality
from sickchill.show.History import History


class HistoryTests(unittest.TestCase):
    """
    Test history
    """

    def test_get_actions(self):
        """
        Tests whether or not the different kinds of actions an episode can have are returned correctly
        """
        test_cases = {
            None: [],
            "": [],
            "wrong": [],
            "downloaded": Quality.DOWNLOADED,
            "Downloaded": Quality.DOWNLOADED,
            "snatched": Quality.SNATCHED,
            "Snatched": Quality.SNATCHED,
        }

        unicode_test_cases = {
            "": [],
            "wrong": [],
            "downloaded": Quality.DOWNLOADED,
            "Downloaded": Quality.DOWNLOADED,
            "snatched": Quality.SNATCHED,
            "Snatched": Quality.SNATCHED,
        }

        for tests in test_cases, unicode_test_cases:
            for (action, result) in tests.items():
                assert History._get_actions(action) == result

    def test_get_limit(self):
        """
        Tests the static get limit method which should return the limit on the amount of elements that should be shown/returned
        """
        test_cases = {
            None: 0,
            "": 0,
            "0": 0,
            "5": 5,
            "-5": 0,
            "1.5": 0,
            "-1.5": 0,
            5: 5,
            -5: 0,
            1.5: 1,
            -1.5: 0,
        }

        unicode_test_cases = {
            "": 0,
            "0": 0,
            "5": 5,
            "-5": 0,
            "1.5": 0,
            "-1.5": 0,
        }

        for tests in test_cases, unicode_test_cases:
            for (action, result) in tests.items():
                assert History._get_limit(action) == result


if __name__ == "__main__":
    print("=====> Testing {0}".format(__file__))

    SUITE = unittest.TestLoader().loadTestsFromTestCase(HistoryTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)




from typing import Any, Dict, List

from datetime import datetime
from github.GithubObject import CompletableGithubObject
from github.NamedUser import NamedUser


class GithubApp(CompletableGithubObject):
    def __repr__(self) -> str: ...
    def _initAttributes(self) -> None: ...
    def _useAttributes(self, attributes: Dict[str, Any]) -> None: ...
    @property
    def id(self) -> str: ...
    @property
    def slug(self) -> str: ...
    @property
    def url(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def description(self) -> str: ...
    @property
    def external_url(self) -> str: ...
    @property
    def html_url(self) -> str: ...
    @property
    def created_at(self) -> datetime: ...
    @property
    def updated_at(self) -> datetime: ...
    @property
    def owner(self) -> NamedUser: ...
    @property
    def permissions(self) -> Dict[str, str]: ...
    @property
    def events(self) -> List[str]: ...



# -*- coding: utf-8 -*-
import logging
import operator

import requests

from .. import __short_version__
from ..cache import REFINER_EXPIRATION_TIME, region
from ..video import Episode, Movie

logger = logging.getLogger(__name__)


class OMDBClient(object):
    base_url = 'http://www.omdbapi.com'

    def __init__(self, version=1, session=None, headers=None, timeout=10):
        #: Session for the requests
        self.session = session or requests.Session()
        self.session.timeout = timeout
        self.session.headers.update(headers or {})
        self.session.params['r'] = 'json'
        self.session.params['v'] = version

    def get(self, id=None, title=None, type=None, year=None, plot='short', tomatoes=False):
        # build the params
        params = {}
        if id:
            params['i'] = id
        if title:
            params['t'] = title
        if not params:
            raise ValueError('At least id or title is required')
        params['type'] = type
        params['y'] = year
        params['plot'] = plot
        params['tomatoes'] = tomatoes

        # perform the request
        r = self.session.get(self.base_url, params=params)
        r.raise_for_status()

        # get the response as json
        j = r.json()

        # check response status
        if j['Response'] == 'False':
            return None

        return j

    def search(self, title, type=None, year=None, page=1):
        # build the params
        params = {'s': title, 'type': type, 'y': year, 'page': page}

        # perform the request
        r = self.session.get(self.base_url, params=params)
        r.raise_for_status()

        # get the response as json
        j = r.json()

        # check response status
        if j['Response'] == 'False':
            return None

        return j


user_agent = 'Subliminal/%s' % __short_version__
omdb_client = OMDBClient(headers={'User-Agent': user_agent})


@region.cache_on_arguments(expiration_time=REFINER_EXPIRATION_TIME)
def search(title, type, year):
    results = omdb_client.search(title, type, year)
    if not results:
        return None

    # fetch all paginated results
    all_results = results['Search']
    total_results = int(results['totalResults'])
    page = 1
    while total_results > page * 10:
        page += 1
        results = omdb_client.search(title, type, year, page=page)
        all_results.extend(results['Search'])

    return all_results


def refine(video, apikey=None, **kwargs):
    """Refine a video by searching `OMDb API <http://omdbapi.com/>`_.

    Several :class:`~subliminal.video.Episode` attributes can be found:

      * :attr:`~subliminal.video.Episode.series`
      * :attr:`~subliminal.video.Episode.year`
      * :attr:`~subliminal.video.Episode.series_imdb_id`

    Similarly, for a :class:`~subliminal.video.Movie`:

      * :attr:`~subliminal.video.Movie.title`
      * :attr:`~subliminal.video.Movie.year`
      * :attr:`~subliminal.video.Video.imdb_id`

    """
    if not apikey:
        logger.warning('No apikey. Skipping omdb refiner.')
        return

    omdb_client.session.params['apikey'] = apikey

    if isinstance(video, Episode):
        # exit if the information is complete
        if video.series_imdb_id:
            logger.debug('No need to search')
            return

        # search the series
        results = search(video.series, 'series', video.year)
        if not results:
            logger.warning('No results for series')
            return
        logger.debug('Found %d results', len(results))

        # filter the results
        results = [r for r in results if video.matches(r['Title'])]
        if not results:
            logger.warning('No matching series found')
            return

        # process the results
        found = False
        for result in sorted(results, key=operator.itemgetter('Year')):
            if video.original_series and video.year is None:
                logger.debug('Found result for original series without year')
                found = True
                break
            if video.year == int(result['Year'].split(u'\u2013')[0]):
                logger.debug('Found result with matching year')
                found = True
                break

        if not found:
            logger.warning('No matching series found')
            return

        # add series information
        logger.debug('Found series %r', result)
        video.series = result['Title']
        video.year = int(result['Year'].split(u'\u2013')[0])
        video.series_imdb_id = result['imdbID']

    elif isinstance(video, Movie):
        # exit if the information is complete
        if video.imdb_id:
            return

        # search the movie
        results = search(video.title, 'movie', video.year)
        if not results:
            logger.warning('No results for movie')
            return
        logger.debug('Found %d results', len(results))

        # filter the results
        results = [r for r in results if video.matches(r['Title'])]
        if not results:
            logger.warning('No matching movie found')
            return

        # process the results
        found = False
        for result in results:
            if video.year is None:
                logger.debug('Found result for movie without year')
                found = True
                break
            if video.year == int(result['Year']):
                logger.debug('Found result with matching year')
                found = True
                break

        if not found:
            logger.warning('No matching movie found')
            return

        # add movie information
        logger.debug('Found movie %r', result)
        video.title = result['Title']
        video.year = int(result['Year'].split(u'\u2013')[0])
        video.imdb_id = result['imdbID']



"""
Tests for SickChill system
"""

import unittest

from .test_restart import RestartTests
from .test_shutdown import ShutdownTests

if __name__ == "__main__":
    print('=====> Running all test in "sickchill_tests.system" <=====')

    TEST_CLASSES = [
        RestartTests,
        ShutdownTests,
    ]

    for test_class in TEST_CLASSES:
        SUITE = unittest.TestLoader().loadTestsFromTestCase(test_class)
        unittest.TextTestRunner(verbosity=2).run(SUITE)



class FanartError(Exception):
    def __str__(self):
        return ', '.join(map(str, self.args))

    def __repr__(self):
        name = self.__class__.__name__
        return '%s%r' % (name, self.args)


class ResponseFanartError(FanartError):
    pass


class RequestFanartError(FanartError):
    pass



from os import environ
import logging
from pkg_resources import parse_version
from time import sleep
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning

try:  # python 3
    from collections.abc import Iterable
    from urllib.parse import urlparse
except ImportError:  # python 2
    from collections import Iterable
    from urlparse import urlparse

import requests
import six

from qbittorrentapi.decorators import login_required
from qbittorrentapi.exceptions import APIConnectionError
from qbittorrentapi.exceptions import HTTPError
from qbittorrentapi.exceptions import HTTP5XXError
from qbittorrentapi.exceptions import LoginFailed
from qbittorrentapi.exceptions import MissingRequiredParameters400Error
from qbittorrentapi.exceptions import InvalidRequest400Error
from qbittorrentapi.exceptions import Unauthorized401Error
from qbittorrentapi.exceptions import Forbidden403Error
from qbittorrentapi.exceptions import NotFound404Error
from qbittorrentapi.exceptions import Conflict409Error
from qbittorrentapi.exceptions import UnsupportedMediaType415Error
from qbittorrentapi.exceptions import InternalServerError500Error
from qbittorrentapi.definitions import APINames

logger = logging.getLogger(__name__)


class Request(object):
    """Facilitates HTTP requests to qBittorrent."""

    def __init__(self, host='', port=None, username=None, password=None, **kwargs):
        self.host = host
        self.port = port
        self.username = username or ''
        self._password = password or ''

        # defaults that should not change
        self._API_URL_BASE_PATH = 'api'
        self._API_URL_API_VERSION = 'v2'

        # state, context, and caching variables
        #   These variables are deleted if the connection to qBittorrent is reset
        #   or a new login is required. All of these (except the SID cookie) should
        #   be reset in _initialize_connection().
        self._SID = None  # authorization cookie
        self._cached_web_api_version = None
        self._application = None
        self._transfer = None
        self._torrents = None
        self._torrent_categories = None
        self._torrent_tags = None
        self._log = None
        self._sync = None
        self._rss = None
        self._search = None
        self._API_URL_BASE = None

        # Configuration variables
        self._VERIFY_WEBUI_CERTIFICATE = kwargs.pop('VERIFY_WEBUI_CERTIFICATE', True)
        self._RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS = kwargs.pop(
            'RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS',
            kwargs.pop('RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS', False))
        self._VERBOSE_RESPONSE_LOGGING = kwargs.pop('VERBOSE_RESPONSE_LOGGING', False)
        self._PRINT_STACK_FOR_EACH_REQUEST = kwargs.pop('PRINT_STACK_FOR_EACH_REQUEST', False)
        self._SIMPLE_RESPONSES = kwargs.pop('SIMPLE_RESPONSES', False)

        if kwargs.pop('DISABLE_LOGGING_DEBUG_OUTPUT', False):
            logging.getLogger('qbittorrentapi').setLevel(logging.INFO)
            logging.getLogger('requests').setLevel(logging.INFO)
            logging.getLogger('urllib3').setLevel(logging.INFO)

        # Environment variables have lowest priority
        if self.host == '' and environ.get('PYTHON_QBITTORRENTAPI_HOST') is not None:
            logger.debug('Using PYTHON_QBITTORRENTAPI_HOST env variable for qBittorrent hostname')
            self.host = environ['PYTHON_QBITTORRENTAPI_HOST']
        if self.username == '' and environ.get('PYTHON_QBITTORRENTAPI_USERNAME') is not None:
            logger.debug('Using PYTHON_QBITTORRENTAPI_USERNAME env variable for username')
            self.username = environ['PYTHON_QBITTORRENTAPI_USERNAME']

        if self._password == '' and environ.get('PYTHON_QBITTORRENTAPI_PASSWORD') is not None:
            logger.debug('Using PYTHON_QBITTORRENTAPI_PASSWORD env variable for password')
            self._password = environ['PYTHON_QBITTORRENTAPI_PASSWORD']

        if self._VERIFY_WEBUI_CERTIFICATE is True and environ.get('PYTHON_QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE') is not None:
            self._VERIFY_WEBUI_CERTIFICATE = False

        # Mocking variables until better unit testing exists
        self._MOCK_WEB_API_VERSION = kwargs.pop('MOCK_WEB_API_VERSION', None)

    ########################################
    # Authorization Endpoints
    ########################################
    @property
    def is_logged_in(self):
        return bool(self._SID)

    def auth_log_in(self, username=None, password=None):
        """
        Log in to qBittorrent host.

        :raises LoginFailed: if credentials failed to log in
        :raises Forbidden403Error: if user user is banned...or not logged in

        :param username: user name for qBittorrent client
        :param password: password for qBittorrent client
        :return: None
        """
        if username:
            self.username = username or ''
            self._password = password or ''

        try:
            self._initialize_context()
            response = self._post(_name=APINames.Authorization,
                                  _method='login',
                                  data={'username': self.username, 'password': self._password})
            self._SID = response.cookies['SID']
        except KeyError:
            logger.debug('Login failed for user "%s"' % self.username)
            raise self._suppress_context(LoginFailed('Login authorization failed for user "%s"' % self.username))
        else:
            logger.debug('Login successful for user "%s"' % self.username)
            logger.debug('SID: %s' % self._SID)

    @login_required
    def auth_log_out(self, **kwargs):
        """End session with qBittorrent."""
        self._get(_name=APINames.Authorization, _method='logout', **kwargs)

    ########################################
    # Helpers
    ########################################
    @classmethod
    def _list2string(cls, input_list=None, delimiter="|"):
        """
        Convert entries in a list to a concatenated string

        :param input_list: list to convert
        :param delimiter: delimiter for concatenation
        :return: if input is a list, concatenated string...else whatever the input was
        """
        if not isinstance(input_list, six.string_types) and isinstance(input_list, Iterable):
            return delimiter.join(map(str, input_list))
        return input_list

    @classmethod
    def _suppress_context(cls, exc):
        """
        This is used to mask an exception with another one.

        For instance, below, the divide by zero error is masked by the CustomException.
            try:
                1/0
            except ZeroDivisionError:
                raise suppress_context(CustomException())

        Note: In python 3, the last line would simply be raise CustomException() from None
        :param exc: new Exception that will be raised
        :return: Exception to be raised
        """
        exc.__cause__ = None
        return exc

    @classmethod
    def _is_version_less_than(cls, ver1, ver2, lteq=True):
        """
        Determine if ver1 is equal to or later than ver2.

        Note: changes need to be reflected in decorators._is_version_less_than as well

        :param ver1: version to check
        :param ver2: current version of application
        :param lteq: True for Less Than or Equals; False for just Less Than
        :return: True or False
        """
        if lteq:
            return parse_version(ver1) <= parse_version(ver2)
        return parse_version(ver1) < parse_version(ver2)

    ########################################
    # HTTP Requests
    ########################################
    def _initialize_context(self):
        """Reset context. This is necessary when the auth cookie needs to be replaced."""
        # cache to avoid perf hit from version checking certain endpoints
        self._cached_web_api_version = None

        # reset URL so the full URL is derived again (primarily allows for switching scheme for WebUI: HTTP <-> HTTPS)
        self._API_URL_BASE = None

        # reinitialize interaction layers
        self._application = None
        self._transfer = None
        self._torrents = None
        self._torrent_categories = None
        self._torrent_tags = None
        self._log = None
        self._sync = None
        self._rss = None
        self._search = None

    def _get(self, _name=APINames.EMPTY, _method='', **kwargs):
        return self._request_wrapper(http_method='get', api_name=_name, api_method=_method, **kwargs)

    def _post(self, _name=APINames.EMPTY, _method='', **kwargs):
        return self._request_wrapper(http_method='post', api_name=_name, api_method=_method, **kwargs)

    def _request_wrapper(self, _retries=2, _retry_backoff_factor=.3, **kwargs):
        """
        Wrapper to manage requests retries.
        This should retry at least twice to account for the Web API switching from HTTP to HTTPS.
        During the second attempt, the URL is rebuilt using HTTP or HTTPS as appropriate.
        """
        max_retries = _retries if _retries > 1 else 2
        for retry in range(0, (max_retries + 1)):
            try:
                return self._request(**kwargs)
            except HTTPError as e:
                # retry the request for HTTP 500 statuses, raise immediately for everything else (e.g. 4XX statuses)
                if not isinstance(e, HTTP5XXError) or retry >= max_retries:
                    raise
            except Exception as e:
                if retry >= max_retries:
                    error_prologue = 'Failed to connect to qBittorrent. '
                    error_messages = {
                        requests.exceptions.SSLError:
                            'This is likely due to using an untrusted certificate (likely self-signed) '
                            'for HTTPS qBittorrent WebUI. To suppress this error (and skip certificate '
                            'verification consequently exposing the HTTPS connection to man-in-the-middle '
                            'attacks), set VERIFY_WEBUI_CERTIFICATE=False when instantiating Client or set '
                            'environment variable PYTHON_QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE '
                            'to a non-null value. SSL Error: %s' % repr(e),
                        requests.exceptions.HTTPError: 'Invalid HTTP Response: %s' % repr(e),
                        requests.exceptions.TooManyRedirects: 'Too many redirects: %s' % repr(e),
                        requests.exceptions.ConnectionError: 'Connection Error: %s' % repr(e),
                        requests.exceptions.Timeout: 'Timeout Error: %s' % repr(e),
                        requests.exceptions.RequestException: 'Requests Error: %s' % repr(e)
                    }
                    error_message = error_prologue + error_messages.get(type(e), 'Unknown Error: %s' % repr(e))
                    logger.debug(error_message)
                    response = getattr(e, 'response', None)
                    raise APIConnectionError(error_message, response=response)

            # back off on attempting each subsequent retry. first retry is always immediate.
            # if the backoff factor is 0.1, then sleep() will sleep for [0.0s, 0.2s, 0.4s, 0.8s, ...] between retries.
            if retry > 0:
                backoff_time = _retry_backoff_factor * (2 ** ((retry + 1) - 1))
                sleep(backoff_time if backoff_time < 30 else 30)
            logger.debug('Retry attempt %d' % (retry+1))
            self._initialize_context()

    def _request(self, http_method, api_name, api_method,
                 data=None, params=None, files=None, headers=None, requests_params=None, **kwargs):
        _ = kwargs.pop('SIMPLE_RESPONSES', kwargs.pop('SIMPLE_RESPONSE', False))  # ensure SIMPLE_RESPONSE(S) isn't sent

        if isinstance(api_name, APINames):
            api_name = api_name.value
        api_path_list = (self._API_URL_BASE_PATH, self._API_URL_API_VERSION, api_name, api_method)
        url = self._build_url(base_url=self._API_URL_BASE,
                              host=self.host,
                              port=self.port,
                              api_path_list=api_path_list)

        # preserve URL without the path so we don't have to rebuild it next time
        self._API_URL_BASE = url._replace(path='')

        # mechanism to send additional arguments to Requests for individual API calls
        requests_params = requests_params or dict()

        # support for custom params to API
        data = data or dict()
        params = params or dict()
        files = files or dict()
        if http_method == 'get':
            params.update(kwargs)
        if http_method == 'post':
            data.update(kwargs)

        # set up headers
        headers = headers or dict()
        headers['Referer'] = self._API_URL_BASE.geturl()
        headers['Origin'] = self._API_URL_BASE.geturl()
        # send Content-Length zero for empty POSTs
        # Requests will not send Content-Length if data is empty
        if http_method == 'post' and not any(filter(None, data.values())):
            headers['Content-Length'] = '0'

        # include the SID auth cookie unless we're trying to log in and get a SID
        cookies = {'SID': self._SID if 'auth/login' not in url.path else ''}

        # turn off console-printed warnings about SSL certificate issues (e.g. untrusted since it is self-signed)
        if not self._VERIFY_WEBUI_CERTIFICATE:
            disable_warnings(InsecureRequestWarning)

        response = requests.request(method=http_method,
                                    url=url.geturl(),
                                    headers=headers,
                                    cookies=cookies,
                                    verify=self._VERIFY_WEBUI_CERTIFICATE,
                                    data=data,
                                    params=params,
                                    files=files,
                                    **requests_params)

        self.verbose_logging(http_method, response, url)
        self.handle_error_responses(data, params, response)
        return response

    @staticmethod
    def handle_error_responses(data, params, response):
        """Raise proper exception if qBittorrent returns Error HTTP Status."""
        if response.status_code < 400:
            # short circuit for non-error statuses
            return
        elif response.status_code == 400:
            # Returned for malformed requests such as missing or invalid parameters.
            #
            # If an error_message isn't returned, qBittorrent didn't receive all required parameters.
            # APIErrorType::BadParams
            if response.text == '':
                raise MissingRequiredParameters400Error()
            raise InvalidRequest400Error(response.text)

        elif response.status_code == 401:
            # Primarily reserved for XSS and host header issues. Is also
            raise Unauthorized401Error(response.text)

        elif response.status_code == 403:
            # Not logged in or calling an API method that isn't public
            # APIErrorType::AccessDenied
            raise Forbidden403Error(response.text)

        elif response.status_code == 404:
            # API method doesn't exist or more likely, torrent not found
            # APIErrorType::NotFound
            error_message = response.text
            if error_message == '':
                error_torrent_hash = ''
                if data:
                    error_torrent_hash = data.get('hash', error_torrent_hash)
                    error_torrent_hash = data.get('hashes', error_torrent_hash)
                if params and error_torrent_hash == '':
                    error_torrent_hash = params.get('hash', error_torrent_hash)
                    error_torrent_hash = params.get('hashes', error_torrent_hash)
                if error_torrent_hash:
                    error_message = 'Torrent hash(es): %s' % error_torrent_hash
            raise NotFound404Error(error_message)

        elif response.status_code == 409:
            # APIErrorType::Conflict
            raise Conflict409Error(response.text)

        elif response.status_code == 415:
            # APIErrorType::BadData
            raise UnsupportedMediaType415Error(response.text)

        elif response.status_code >= 500:
            raise InternalServerError500Error(response.text)

        elif response.status_code >= 400:
            # Unaccounted for errors from API
            raise HTTPError(response.text)

    def verbose_logging(self, http_method, response, url):
        """Log verbose information about request. Can be useful during development."""
        if self._VERBOSE_RESPONSE_LOGGING:
            resp_logger = logger.debug
            max_text_length_to_log = 254
            if response.status_code != 200:
                max_text_length_to_log = 10000  # log as much as possible in a error condition

            resp_logger('Request URL: (%s) %s' % (http_method.upper(), response.url))
            if str(response.request.body) not in ('None', '') and 'auth/login' not in url.path:
                body_len = max_text_length_to_log if len(response.request.body) > max_text_length_to_log else len(response.request.body)
                resp_logger('Request body: %s%s' % (response.request.body[:body_len], '...<truncated>' if body_len >= 80 else ''))

            resp_logger('Response status: %s (%s)' % (response.status_code, response.reason))
            if response.text:
                text_len = max_text_length_to_log if len(response.text) > max_text_length_to_log else len(response.text)
                resp_logger(
                    'Response text: %s%s' % (response.text[:text_len], '...<truncated>' if text_len >= 80 else ''))
        if self._PRINT_STACK_FOR_EACH_REQUEST:
            from traceback import print_stack
            print_stack()

    @staticmethod
    def _build_url(base_url=None, host='', port=None, api_path_list=None):
        """
        Create a fully qualified URL (minus query parameters that Requests will add later).

        Supports detecting whether HTTPS is enabled for WebUI.

        :param base_url: if the URL was already built, this is the base URL
        :param host: user provided hostname for WebUI
        :param api_path_list: list of strings for API endpoint path (e.g. ['api', 'v2', 'app', 'version'])
        :return: full URL for WebUI API endpoint
        """
        # build full URL if it's the first time we're here
        if base_url is None:
            if not host.lower().startswith(('http:', 'https:', '//')):
                host = '//' + host
            base_url = urlparse(url=host)
            # force scheme to HTTP even if host was provided with HTTPS scheme
            base_url = base_url._replace(scheme='http')
            # add port number if host doesn't contain one
            if port is not None and not isinstance(base_url.port, int):
                base_url = base_url._replace(netloc='%s:%s' % (base_url.netloc, port))

            # detect whether Web API is configured for HTTP or HTTPS
            logger.debug('Detecting scheme for URL...')
            try:
                r = requests.head(base_url.geturl(), allow_redirects=True)
                # if WebUI eventually supports sending a redirect from HTTP to HTTPS then
                # Requests will automatically provide a URL using HTTPS.
                # For instance, the URL returned below will use the HTTPS scheme.
                #  >>> requests.head('http://grc.com', allow_redirects=True).url
                scheme = urlparse(r.url).scheme
            except requests.exceptions.RequestException:
                # qBittorrent will reject the connection if WebUI is configured for HTTPS.
                # If something else caused this exception, we'll properly handle that
                # later during the actual API request.
                scheme = 'https'

            # use detected scheme
            logger.debug('Using %s scheme' % scheme.upper())
            base_url = base_url._replace(scheme=scheme)

            logger.debug('Base URL: %s' % base_url.geturl())

        # add the full API path to complete the URL
        return base_url._replace(path='/'.join(map(lambda s: s.strip('/'), map(str, api_path_list))))



"""
Create a test database for testing.

Methods:
    create_test_log_folder
    create_test_cache_folder
    setup_test_db
    teardown_test_db
    setup_test_episode_file
    teardown_test_episode_file
    setup_test_show_dir
    teardown_test_show_dir

Classes:
    SickChillTestDBCase
    TestDBConnection
    TestCacheDBConnection
"""

import os
import os.path
import shutil
import unittest

import pytest
from configobj import ConfigObj

import sickchill.logger
import sickchill.oldbeard.config
import sickchill.oldbeard.tvcache
import sickchill.start
from sickchill import settings
from sickchill.oldbeard import db, providers
from sickchill.oldbeard.databases import cache, failed, main
from sickchill.show.indexers import ShowIndexer
from sickchill.tv import TVEpisode, TVShow

# =================
#  test globals
# =================
TEST_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DB_NAME = "sickchill.db"
TEST_CACHE_DB_NAME = "cache.db"
TEST_FAILED_DB_NAME = "failed.db"

SHOW_NAME = "show name"
SEASON = 4
EPISODE = 2
FILENAME = "show name - s0" + str(SEASON) + "e0" + str(EPISODE) + ".mkv"
FILE_DIR = os.path.join(TEST_DIR, SHOW_NAME)
FILE_PATH = os.path.join(FILE_DIR, FILENAME)
SHOW_DIR = os.path.join(TEST_DIR, SHOW_NAME + " final")
PROCESSING_DIR = os.path.join(TEST_DIR, "Downloads")
NUM_SEASONS = 5
EPISODES_PER_SEASON = 20

# =================
#  prepare env functions
# =================


def create_test_log_folder():
    """
    Create a log folder for test logs.
    """
    if not os.path.isdir(settings.LOG_DIR):
        os.mkdir(settings.LOG_DIR)


def create_test_cache_folder():
    """
    Create a cache folder for caching tests.
    """
    if not os.path.isdir(settings.CACHE_DIR):
        os.mkdir(settings.CACHE_DIR)


# call env functions at appropriate time during SickChill var setup

# =================
#  SickChill globals
# =================


settings.showList = []
settings.QUALITY_DEFAULT = 4  # hdtv
settings.SEASON_FOLDERS_DEFAULT = 0

settings.NAMING_PATTERN = ""
settings.NAMING_ABD_PATTERN = ""
settings.NAMING_SPORTS_PATTERN = ""
settings.NAMING_MULTI_EP = 1

settings.TV_DOWNLOAD_DIR = PROCESSING_DIR

# settings.PROVIDER_ORDER = ["sick_beard_index"]
settings.providerList = providers.makeProviderList()

settings.DATA_DIR = TEST_DIR
settings.CONFIG_FILE = os.path.join(settings.DATA_DIR, "config.ini")
settings.CFG = ConfigObj(settings.CONFIG_FILE, encoding="UTF-8", indent_type="  ")
settings.GUI_NAME = "slick"

settings.BRANCH = sickchill.oldbeard.config.check_setting_str(settings.CFG, "General", "branch")
settings.CUR_COMMIT_HASH = sickchill.oldbeard.config.check_setting_str(settings.CFG, "General", "cur_commit_hash")
settings.GIT_USERNAME = sickchill.oldbeard.config.check_setting_str(settings.CFG, "General", "git_username")
settings.GIT_TOKEN = sickchill.oldbeard.config.check_setting_str(settings.CFG, "General", "git_token_password", censor_log=True)

settings.LOG_DIR = os.path.join(TEST_DIR, "Logs")
sickchill.logger.log_file = os.path.join(settings.LOG_DIR, "test_sickchill.log")
create_test_log_folder()

settings.CACHE_DIR = os.path.join(TEST_DIR, "cache")
create_test_cache_folder()

sickchill.logger.init_logging(False, True)

sickchill.indexer = ShowIndexer()

# =================
#  dummy functions
# =================
def _dummy_save_config():
    """
    Override the SickChill save_config which gets called during a db upgrade.

    :return: True
    """
    return True


# this overrides the SickChill save_config which gets called during a db upgrade
# this might be considered a hack
sickchill.start.save_config = _dummy_save_config


def _fake_specify_ep(self, season, episode):
    """
    Override contact to TVDB indexer.

    :param self: ...not used
    :param season: Season to search for  ...not used
    :param episode: Episode to search for  ...not used
    """
    _ = self, season, episode  # throw away unused variables


# the real one tries to contact TVDB just stop it from getting more info on the ep
TVEpisode.specifyEpisode = _fake_specify_ep


# =================
#  test classes
# =================
class SickChillTestDBCase(unittest.TestCase):
    """
    Superclass for testing the database.

    Methods:
        setUp
        tearDown
    """

    def setUp(self):
        settings.showList = []
        setup_test_db()
        setup_test_episode_file()
        setup_test_show_dir()

    def tearDown(self):
        settings.showList = []
        teardown_test_db()
        teardown_test_episode_file()
        teardown_test_show_dir()


class SickChillTestPostProcessorCase(unittest.TestCase):
    """
    Superclass for testing the database.

    Methods:
        setUp
        tearDown
    """

    def setUp(self):
        settings.showList = []
        setup_test_db()
        setup_test_episode_file()
        setup_test_show_dir()
        setup_test_processing_dir()

        self.show = TVShow(1, 1, "en")
        self.show.name = SHOW_NAME
        self.show.location = FILE_DIR
        self.show.imdb_info = {"indexer_id": self.show.indexerid, "imdb_id": "tt000000"}

        self.show.episodes = {}
        for season in range(1, NUM_SEASONS):
            self.show.episodes[season] = {}
            for episode in range(1, EPISODES_PER_SEASON):
                if season == SEASON and episode == EPISODE:
                    episode = TVEpisode(self.show, season, episode, ep_file=FILE_PATH)
                else:
                    episode = TVEpisode(self.show, season, episode)
                self.show.episodes[season][episode] = episode
                episode.saveToDB()

        self.show.saveToDB()
        settings.showList = [self.show]

    def tearDown(self):
        settings.showList = []
        self.show.deleteShow(True)
        teardown_test_db()
        teardown_test_episode_file()
        teardown_test_show_dir()
        teardown_test_processing_dir()


class TestDBConnection(db.DBConnection, object):
    """
    Test connecting to the database.
    """

    __test__ = False

    def __init__(self, filename=TEST_DB_NAME, suffix=None, row_type=None):
        if TEST_DIR not in filename:
            filename = os.path.join(TEST_DIR, filename)
        super().__init__(filename=filename, suffix=suffix, row_type=row_type)


class TestCacheDBConnection(TestDBConnection, object):
    """
    Test connecting to the cache database.
    """

    __test__ = False

    def __init__(self, filename=TEST_CACHE_DB_NAME, suffix=None, row_type="dict"):
        if TEST_DIR not in filename:
            filename = os.path.join(TEST_DIR, filename)
        super().__init__(filename=filename, suffix=suffix, row_type=row_type)


# this will override the normal db connection
sickchill.oldbeard.db.DBConnection = TestDBConnection
sickchill.oldbeard.tvcache.CacheDBConnection = TestCacheDBConnection


# =================
#  test functions
# =================
def setup_test_db():
    """
    Set up the test databases.
    """
    # Upgrade the db to the latest version.
    # upgrading the db
    db.upgrade_database(db.DBConnection(), main.InitialSchema)

    # fix up any db problems
    db.sanity_check_database(db.DBConnection(), main.MainSanityCheck)

    # and for cache.db too
    db.upgrade_database(db.DBConnection("cache.db"), cache.InitialSchema)

    # and for failed.db too
    db.upgrade_database(db.DBConnection("failed.db"), failed.InitialSchema)


def teardown_test_db():
    """
    Tear down the test database.
    """
    from sickchill.oldbeard.db import db_cons

    for connection in db_cons:
        db_cons[connection].commit()
    #     db_cons[connection].close()
    #
    # for current_db in [ TEST_DB_NAME, TEST_CACHE_DB_NAME, TEST_FAILED_DB_NAME ]:
    #    filename = os.path.join(TEST_DIR, current_db)
    #    if os.path.exists(filename):
    #        try:
    #            os.remove(filename)
    #        except Exception as e:
    #            print('ERROR: Failed to remove ' + filename)
    #            print(Exception(e))


def setup_test_episode_file():
    """
    Create a test episode directory with a test episode in it.
    """
    if not os.path.exists(FILE_DIR):
        os.makedirs(FILE_DIR)

    try:
        with open(FILE_PATH, "w") as ep_file:
            ep_file.write("foo bar")
            ep_file.flush()

    # Catching too general exception
    except Exception:
        print("Unable to set up test episode")
        raise


def setup_test_processing_dir():
    if not os.path.exists(PROCESSING_DIR):
        os.makedirs(PROCESSING_DIR)

    for season in range(1, NUM_SEASONS):
        for episode in range(11, EPISODES_PER_SEASON):
            path = os.path.join(
                PROCESSING_DIR, "{show_name}.S0{season}E{episode}.HDTV.x264.[SickChill].mkv".format(show_name=SHOW_NAME, season=season, episode=episode)
            )
            with open(path, "w") as ep_file:
                ep_file.write("foo bar")
                ep_file.flush()


def teardown_test_processing_dir():
    if os.path.exists(PROCESSING_DIR):
        shutil.rmtree(PROCESSING_DIR)


def teardown_test_episode_file():
    """
    Remove the test episode.
    """
    if os.path.exists(FILE_DIR):
        shutil.rmtree(FILE_DIR)


def setup_test_show_dir():
    """
    Create a test show directory.
    """
    if not os.path.exists(SHOW_DIR):
        os.makedirs(SHOW_DIR)


def teardown_test_show_dir():
    """
    Remove the test show.
    """
    if os.path.exists(SHOW_DIR):
        shutil.rmtree(SHOW_DIR)


@pytest.fixture(scope="module")
def vcr_cassette_dir(request):
    # Put all cassettes in vhs/{module}/{test}.yaml
    return os.path.join(os.path.dirname(request.module.__file__), "cassettes")


@pytest.fixture()
def vcr_cassette_name(request):
    """Name of the VCR cassette"""
    return request.cls.provider.get_id() + ".yaml"


def patch_open(open_func, files):
    def open_patched(path, mode='r', buffering=-1, encoding=None,
                     errors=None, newline=None, closefd=True,
                     opener=None):
        if 'w' in mode and not os.path.isfile(path):
            files.append(path)
        return open_func(path, mode=mode, buffering=buffering,
                         encoding=encoding, errors=errors,
                         newline=newline, closefd=closefd,
                         opener=opener)
    return open_patched


@pytest.fixture(autouse=True, scope="session")
def cleanup_files():
    # yield
    for file in [ "tests/sickchill.db", "tests/cache.db", "tests/failed.db"]:
        if os.path.exists(file):
            print(f"Removing {file}")
            os.remove(file)



# Copyright 2010-2020 Kurt McKee <contactme@kurtmckee.org>
# Copyright 2002-2008 Mark Pilgrim
# All rights reserved.
#
# This file is a part of feedparser.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 'AS IS'
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import re

__all__ = [
    '_SGML_AVAILABLE',
    'sgmllib',
    'charref',
    'tagfind',
    'attrfind',
    'entityref',
    'incomplete',
    'interesting',
    'shorttag',
    'shorttagopen',
    'starttagopen',
    'endbracket',
]

# sgmllib is not available by default in Python 3; if the end user doesn't have
# it available then we'll lose illformed XML parsing and content sanitizing
try:
    import sgmllib
except ImportError:
    # This is probably Python 3, which doesn't include sgmllib anymore
    _SGML_AVAILABLE = 0

    # Mock sgmllib enough to allow subclassing later on
    class sgmllib(object):
        SGMLParseError = EnvironmentError

        class SGMLParser(object):
            lasttag = None
            rawdata = None

            def close(self):
                pass

            def feed(self, data):
                pass

            def goahead(self, i):
                pass

            def parse_declaration(self, i):
                pass

            def parse_starttag(self, i):
                pass

            def reset(self):
                pass

else:
    _SGML_AVAILABLE = 1

    # sgmllib defines a number of module-level regular expressions that are
    # insufficient for the XML parsing feedparser needs. Rather than modify
    # the variables directly in sgmllib, they're defined here using the same
    # names, and the compiled code objects of several sgmllib.SGMLParser
    # methods are copied into _BaseHTMLProcessor so that they execute in
    # feedparser's scope instead of sgmllib's scope.
    charref = re.compile(r'&#(\d+|[xX][0-9a-fA-F]+);')
    tagfind = re.compile(r'[a-zA-Z][-_.:a-zA-Z0-9]*')
    attrfind = re.compile(
        r"""\s*([a-zA-Z_][-:.a-zA-Z_0-9]*)[$]?(\s*=\s*"""
        r"""('[^']*'|"[^"]*"|[][\-a-zA-Z0-9./,:;+*%?!&$()_#=~'"@]*))?"""
    )

    # Unfortunately, these must be copied over to prevent NameError exceptions
    entityref = sgmllib.entityref
    incomplete = sgmllib.incomplete
    interesting = sgmllib.interesting
    shorttag = sgmllib.shorttag
    shorttagopen = sgmllib.shorttagopen
    starttagopen = sgmllib.starttagopen


    class _EndBracketRegEx:
        def __init__(self):
            # Overriding the built-in sgmllib.endbracket regex allows the
            # parser to find angle brackets embedded in element attributes.
            self.endbracket = re.compile(
                r'('
                r"""[^'"<>]"""
                r"""|"[^"]*"(?=>|/|\s|\w+=)"""
                r"""|'[^']*'(?=>|/|\s|\w+=))*(?=[<>])"""
                r"""|.*?(?=[<>]"""
                r')'
            )

        def search(self, target, index=0):
            match = self.endbracket.match(target, index)
            if match is not None:
                # Returning a new object in the calling thread's context
                # resolves a thread-safety.
                return EndBracketMatch(match)
            return None


    class EndBracketMatch:
        def __init__(self, match):
            self.match = match

        def start(self, n):
            return self.match.end(n)


    endbracket = _EndBracketRegEx()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["เมื่อสักครู่นี้", "อีกสักครู่"],
    ["%s วินาทีที่แล้ว", "ใน %s วินาที"],
    ["1 นาทีที่แล้ว", "ใน 1 นาที"],
    ["%s นาทีที่แล้ว", "ใน %s นาที"],
    ["1 ชั่วโมงที่แล้ว", "ใน 1 ชั่วโมง"],
    ["%s ชั่วโมงที่แล้ว", "ใน %s ชั่วโมง"],
    ["1 วันที่แล้ว", "ใน 1 วัน"],
    ["%s วันที่แล้ว", "ใน %s วัน"],
    ["1 อาทิตย์ที่แล้ว", "ใน 1 อาทิตย์"],
    ["%s อาทิตย์ที่แล้ว", "ใน %s อาทิตย์"],
    ["1 เดือนที่แล้ว", "ใน 1 เดือน"],
    ["%s เดือนที่แล้ว", "ใน %s เดือน"],
    ["1 ปีที่แล้ว", "ใน 1 ปี"],
    ["%s ปีที่แล้ว", "ใน %s ปี"]
]
        


#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020-04-17

@author: naglis
'''

base = [
    ["ką tik", "po akimirkos"],
    ["prieš %s sekundę", "po %s sekundės", "prieš %s sekundes", "po %s sekundžių", "prieš %s sekundžių", "po %s sekundžių"],
    ["prieš 1 minutę", "po 1 minutės"],
    ["prieš %s minutę", "po %s minutės", "prieš %s minutes", "po %s minučių", "prieš %s minučių", "po %s minučių"],
    ["prieš 1 valandą", "po 1 valandos"],
    ["prieš %s valandą", "po %s valandos", "prieš %s valandas", "po %s valandų", "prieš %s valandų", "po %s valandų"],
    ["prieš 1 dieną", "po 1 dienos"],
    ["prieš %s dieną", "po %s dienos", "prieš %s dienas", "po %s dienų", "prieš %s dienų", "po %s dienų"],
    ["prieš 1 savaitę", "po 1 savaitės"],
    ["prieš %s savaitę", "po %s savaitės", "prieš %s savaites", "po %s savaičių", "prieš %s savaičių", "po %s savaičių"],
    ["prieš 1 mėnesį", "po 1 mėnesio"],
    ["prieš %s mėnesį", "po %s mėnesio", "prieš %s mėnesius", "po %s mėnesių", "prieš %s mėnesių", "po %s mėnesių"],
    ["prieš 1 metus", "po 1 metų"],
    ["prieš %s metus", "po %s metų", "prieš %s metus", "po %s metų", "prieš %s metų", "po %s metų"],
]


def generate(row, y):
    def formatting(time):
        if row % 2 == 0:
            return base[row][y]

        r_10, r_100 = time % 10, time % 100
        if 10 < r_100 < 20 or r_10 == 0:
            return base[row][y + 4]
        elif r_10 != 1:
            return base[row][y + 2]
        else:
            return base[row][y]

    return formatting


LOCALE = generate



"""
Test tv
"""

import unittest

from sickchill import settings
from sickchill.tv import TVEpisode, TVShow
from tests import conftest


class TVShowTests(conftest.SickChillTestPostProcessorCase):
    """
    Test tv shows
    """

    def setUp(self):
        """
        Set up tests
        """
        super().setUp()

    def test_init_indexerid(self):
        """
        test init indexer id
        """
        self.show.loadFromDB()
        assert self.show.indexerid == 1

    def test_change_indexerid(self):
        """
        test change indexer id
        """
        self.show.loadFromDB()
        assert self.show.indexerid == 1

        self.show.indexerid = 295759
        self.show.saveToDB()
        self.show.loadFromDB()

        assert self.show.indexerid == 295759

    def test_set_name(self):
        """
        test set name
        """
        self.show.name = "newName"

        self.show.saveToDB()
        self.show.loadFromDB()
        assert self.show.name == "newName"
        assert self.show.show_name == "newName"

        self.show.show_name = "show_name"

        self.show.saveToDB()
        self.show.loadFromDB()
        assert self.show.name == "show_name"

    def test_show_with_custom_name(self):
        """
        test set custom name
        """
        self.show.name = "show name"
        self.show.show_name = "show name"
        self.show.custom_name = "newName"
        self.show.saveToDB()
        self.show.loadFromDB()

        assert self.show.show_name == "show name"

        assert self.show.custom_name == "newName"
        assert self.show.name == "newName"


class TVEpisodeTests(conftest.SickChillTestDBCase):
    """
    Test tv episode
    """

    def setUp(self):
        """
        Set up
        """
        super().setUp()
        settings.showList = []

    def test_init_empty_db(self):
        """
        test init empty db
        """
        show = TVShow(1, 1, "en")
        episode = TVEpisode(show, 1, 1)
        episode.name = "asdasdasdajkaj"
        episode.saveToDB()
        episode.loadFromDB(1, 1)
        assert episode.name == "asdasdasdajkaj"


class TVTests(conftest.SickChillTestDBCase):
    """
    Test tv
    """

    def setUp(self):
        """
        Set up
        """
        super().setUp()
        settings.showList = []

    @staticmethod
    def test_get_episode():
        """
        Test get episodes
        """
        show = TVShow(1, 1, "en")
        show.name = "show name"
        show.network = "cbs"
        show.genre = ["crime"]
        show.runtime = 40
        show.status = "Ended"
        show.default_ep_status = "5"
        show.airs = "monday"
        show.startyear = 1987
        show.saveToDB()
        settings.showList = [show]
        # TODO: implement


if __name__ == "__main__":
    print("==================")
    print("STARTING - TV TESTS")
    print("==================")
    print("######################################################################")
    SUITE = unittest.TestLoader().loadTestsFromTestCase(TVShowTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)
    print("######################################################################")
    SUITE = unittest.TestLoader().loadTestsFromTestCase(TVEpisodeTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)
    print("######################################################################")
    SUITE = unittest.TestLoader().loadTestsFromTestCase(TVTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



import datetime
import threading

from .. import logger


class QueuePriorities(object):
    LOW = 10
    NORMAL = 20
    HIGH = 30


class GenericQueue(object):
    def __init__(self):

        self.amActive = False

        self.currentItem = None

        self.queue = []

        self.queue_name = "QUEUE"

        self.min_priority = 0

        self.lock = threading.Lock()

    def __len__(self):
        _len = len(self.queue)
        if self.currentItem:
            _len += 1
        return _len

    def pause(self):
        """Pauses this queue"""
        logger.info("Pausing queue")
        self.min_priority = 999999999999

    def unpause(self):
        """Unpauses this queue"""
        logger.info("Unpausing queue")
        self.min_priority = 0

    def add_item(self, item):
        """
        Adds an item to this queue

        :param item: Queue object to add
        :return: item
        """
        with self.lock:
            item.added = datetime.datetime.now()
            self.queue.append(item)

            return item

    def run(self, force=False):
        """
        Process items in this queue

        :param force: Force queue processing (currently not implemented)
        """
        self.amActive = True

        with self.lock:
            # only start a new task if one isn't already going
            if self.currentItem is None or not self.currentItem.is_alive():

                # if the thread is dead then the current item should be finished
                if self.currentItem:
                    self.currentItem.finish()
                    self.currentItem = None

                # if there's something in the queue then run it in a thread and take it out of the queue
                if self.queue:

                    from functools import cmp_to_key

                    # sort by priority
                    def sorter(x, y):
                        """
                        Sorts by priority descending then time ascending
                        """
                        if x.priority == y.priority:
                            if y.added == x.added:
                                return 0
                            elif y.added < x.added:
                                return 1
                            elif y.added > x.added:
                                return -1
                        else:
                            return y.priority - x.priority

                    self.queue.sort(key=cmp_to_key(sorter))
                    if self.queue[0].priority < self.min_priority:
                        return

                    # launch the queue item in a thread
                    self.currentItem = self.queue.pop(0)
                    self.currentItem.name = self.queue_name + "-" + self.currentItem.name
                    self.currentItem.start()

        self.amActive = False


class QueueItem(threading.Thread):
    def __init__(self, name, action_id=0):
        super().__init__()

        self.name = name.replace(" ", "-").upper()
        self.inProgress = False
        self.priority = QueuePriorities.NORMAL
        self.action_id = action_id
        self.stop = threading.Event()
        self.added = None

    def run(self):
        """Implementing classes should call this"""

        self.inProgress = True

    def finish(self):
        """Implementing Classes should call this"""

        self.inProgress = False

        threading.currentThread().name = self.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["for et øjeblik siden", "om et øjeblik"],
    ["for %s sekunder siden", "om %s sekunder"],
    ["for 1 minut siden", "om 1 minut"],
    ["for %s minutter siden", "om %s minutter"],
    ["for 1 time siden", "om 1 time"],
    ["for %s timer siden", "om %s timer"],
    ["for 1 dag siden", "om 1 dag"],
    ["for %s dage siden", "om %s dage"],
    ["for 1 uge siden", "om 1 uge"],
    ["for %s uger siden", "om %s uger"],
    ["for 1 måned siden", "om 1 måned"],
    ["for %s måneder siden", "om %s måneder"],
    ["for 1 år siden", "om 1 år"],
    ["for %s år siden", "om %s år"]
]
        


#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2016-5-26

@author: hustcc
'''

LOCALE = [
    ["just now", "a while"],
    ["%s seconds ago", "in %s seconds"],
    ["1 minute ago", "in 1 minute"],
    ["%s minutes ago", "in %s minutes"],
    ["1 hour ago", "in 1 hour"],
    ["%s hours ago", "in %s hours"],
    ["1 day ago", "in 1 day"],
    ["%s days ago", "in %s days"],
    ["1 week ago", "in 1 week"],
    ["%s weeks ago", "in %s weeks"],
    ["1 month ago", "in 1 month"],
    ["%s months ago", "in %s months"],
    ["1 year ago", "in 1 year"],
    ["%s years ago", "in %s years"],
]



import datetime
import json
import logging
import threading

import imdb
import tmdbsimple
from imdb.parser.http.piculet import Path, Rule
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from tmdbsimple import movies, search

from . import settings
from .oldbeard.databases import movie
from .oldbeard.db import db_cons, db_full_path, db_locks

logger = logging.getLogger("sickchill.movie")


class MovieList:
    def __init__(self):
        tmdbsimple.API_KEY = settings.TMDB_API_KEY

        self.filename = "movies.db"
        self.full_path = db_full_path(self.filename)

        if self.filename not in db_cons or not db_cons[self.filename]:
            movie.Session.configure(
                bind=create_engine(
                    f"sqlite:///{self.full_path}",
                    echo=True,
                    connect_args={"check_same_thread": False},
                    json_serializer=lambda obj: json.dumps(obj, ensure_ascii=False),
                )
            )
            self.session: Session = movie.Session()
            movie.Base.metadata.create_all(self.session.bind, checkfirst=True)

            db_locks[self.filename] = threading.Lock()
            db_cons[self.filename] = self.session
        else:
            self.session: Session = db_cons[self.filename]

        self.imdb = imdb.IMDb()
        try:
            self.imdb.topBottomProxy.moviemeter100_parser.rules[0].extractor.rules.append(
                Rule(key="cover url", extractor=Path('./td[@class="posterColumn"]/a/img/@src'))
            )
        except:
            pass

    def __iter__(self):
        for item in self.query.all():
            yield item

    def __getitem__(self, pk):
        return self.query.get(pk)

    def __contains__(self, pk):
        try:
            self.__getitem__(pk)
            return True
        except KeyError:
            return False

    @staticmethod
    def search_tmdb(query=None, tmdb_id=None, year=None, language=None, adult=False):
        if tmdb_id:
            results = [movies.Movies(id=tmdb_id)]
        elif query:
            tmdb_kwargs = dict(query=query, year=year, language=language, adult=adult)
            tmdb_kwargs = {key: value for key, value in tmdb_kwargs.items() if value}
            results = search.Search().movie(**tmdb_kwargs)["results"]
        else:
            raise Exception("Query or tmdb id is required!")

        return results

    @staticmethod
    def popular_tmdb(language=None):
        tmdb_kwargs = dict(language=language) if language else dict()
        return movies.Movies().popular(**tmdb_kwargs)["results"]

    def search_imdb(self, query: str = ""):
        return self.imdb.search_movie(title=query)

    def popular_imdb(self):
        return self.imdb.get_popular100_movies()

    def add_from_tmdb(self, tmdb_id: str, language: str = settings.INDEXER_DEFAULT_LANGUAGE):
        logger.debug(f"Adding movie from tmdb with id: {tmdb_id}")
        existing = self.session.query(movie.IndexerData).filter_by(pk=tmdb_id).first()
        if existing:
            logger.debug(f"Movie already existed as {existing.movie.name}")
            return existing.movie

        tmdb_object = tmdbsimple.movies.Movies(id=tmdb_id).info()
        return self.add_from_imdb(tmdb_object["imdb_id"], language, tmdb_primary=True)

    def add_from_imdb(self, imdb_id: str, language: str = settings.INDEXER_DEFAULT_LANGUAGE, tmdb_primary=False):
        if not tmdb_primary:
            logger.debug(f"Adding movie from imdb id: {imdb_id}")

        existing = self.session.query(movie.IndexerData).filter_by(pk=imdb_id).first()
        if existing:
            logger.debug(f"Movie already existed as {existing.movie.name}")
            return existing.movie

        imdb_object = self.imdb.get_movie(imdb_id)
        tmdb_id = tmdbsimple.find.Find(id=imdb_id).info(external_source="imdb_id")["movie_results"][0]["id"]
        tmdb_object = tmdbsimple.movies.Movies(id=tmdb_id).info()

        if tmdb_primary:
            instance = movie.Movie(tmdb_object["title"], year=tmdb_object["release_date"].split("-")[0])
            if imdb_object["title"] and not instance.name:
                instance.name = imdb_object["title"]
        else:
            instance = movie.Movie(imdb_object["title"], year=imdb_object["year"] or tmdb_object)
            if tmdb_object["release_date"] and not instance.year:
                instance.year = tmdb_object["release_date"].split("-")[0]
            if tmdb_object["title"] and not instance.name:
                instance.name = tmdb_object["title"]

        if tmdb_object["release_date"]:
            instance.date = datetime.datetime.strptime(tmdb_object["release_date"], "%Y-%m-%d").date()

        instance.language = tmdb_object["original_language"] or language

        tmdb_data = movie.IndexerData(site="tmdb", data=tmdb_object, pk=tmdb_id)
        imdb_data = movie.IndexerData(site="imdb", data=imdb_object, pk=imdb_id)

        imdb_genres = self.imdb.get_title_genres(imdb_id)["genres"]

        def add_imdb_genres():
            for genre in imdb_genres:
                logger.debug(f"Adding imdb genre {genre}")
                imdb_data.genres.append(movie.Genres(pk=genre))
            instance.indexer_data.append(imdb_data)

        def add_tmdb_genres():
            for genre in tmdb_object["genres"]:
                if genre["name"] not in imdb_genres:
                    logger.debug(f'Adding tmdb genre {genre["name"]}')
                    tmdb_data.genres.append(movie.Genres(pk=genre["name"]))
            instance.indexer_data.append(tmdb_data)

        if tmdb_primary:
            add_tmdb_genres()
            add_imdb_genres()
        else:
            add_imdb_genres()
            add_tmdb_genres()

        self.commit(instance)

        logger.debug(f"Returning instance for {instance.name}")
        return instance

    def commit(self, instance=None):
        logger.debug("Committing")
        if instance:
            self.session.add(instance)
        self.session.flush()
        self.session.commit()

    def delete(self, pk):
        instance = self.query.get(pk)
        if instance:
            self.session.delete(instance)

    @property
    def query(self):
        return self.session.query(movie.Movie)

    def by_slug(self, slug):
        return self.query.filter_by(slug=slug).first()

    def search_providers(self, movie_object: movie.Movie):
        # We should only need to support backlog for movies manually
        # Movies should be returned in the existing RSS searches if we have movies in our list
        strings = movie_object.search_strings()
        for provider in settings.providerList:
            if provider.can_backlog and provider.backlog_enabled and provider.supports_movies:
                results = provider.search(strings)
                for result in results:
                    movie.Result(result=result, provider=provider, movie=movie_object)

            self.commit(movie_object)
            # TODO: Check if we need to break out here and stop hitting providers if we found a good result

    def snatch_movie(self, result: movie.Result):
        pass

    def search_thread(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["orain", "denbora bat barru"],
    ["duela %s segundu", "%s segundu barru"],
    ["duela minutu 1", "minutu 1 barru"],
    ["duela %s minutu", "%s minutu barru"],
    ["duela ordu 1", "ordu 1 barru"],
    ["duela %s ordu", "%s ordu barru"],
    ["duela egun 1", "egun 1 barru"],
    ["duela %s egun", "%s egun barru"],
    ["duela aste 1", "aste 1 barru"],
    ["duela %s aste", "%s aste barru"],
    ["duela hillabete 1", "hillabete 1 barru"],
    ["duela %s hillabete", "%s hillabete barru"],
    ["duela urte 1", "urte 1 barru"],
    ["duela %s urte", "%s urte barru"]
]
        


from unittest import TestCase
import os
from fanart.items import LeafItem
from httpretty import httprettified, HTTPretty
from fanart.tests import LOCALDIR


class LeafItemTestCase(TestCase):
    def setUp(self):
        self.leaf = LeafItem(id=11977, likes=2, url='http://test.tv/50x50.txt')

    def test_str(self):
        self.assertEqual(str(self.leaf), 'http://test.tv/50x50.txt')

    @httprettified
    def test_content(self):
        with open(os.path.join(LOCALDIR, 'response/50x50.png'), 'rb') as fp:
            body = fp.read()
        HTTPretty.register_uri(
            HTTPretty.GET,
            'http://test.tv/50x50.txt',
            body=body
        )
        self.assertEqual(self.leaf.content(), body)
        self.assertEqual(len(HTTPretty.latest_requests), 1)
        self.assertEqual(self.leaf.content(), body)  # Cached
        self.assertEqual(len(HTTPretty.latest_requests), 1)



"""
Tests for SickChill
"""

import unittest

from . import helper, providers, show, system

if __name__ == "__main__":
    print('=====> Running all test in "sickchill_tests" <=====')

    TEST_MODULES = [
        helper,
        providers,
        show,
        system,
    ]

    for test_module in TEST_MODULES:
        SUITE = unittest.TestLoader().loadTestsFromModule(test_module)
        unittest.TextTestRunner(verbosity=2).run(SUITE)



from sickchill import settings


class DelugeBase(object):
    @staticmethod
    def make_options(result):
        options = {}

        if settings.TORRENT_PATH or settings.TORRENT_PATH_INCOMPLETE:
            options.update(
                {
                    "download_location": settings.TORRENT_PATH_INCOMPLETE or settings.TORRENT_PATH,
                    "move_completed": bool(settings.TORRENT_PATH_INCOMPLETE),
                }
            )

        if settings.TORRENT_PATH and settings.TORRENT_PATH_INCOMPLETE:
            options.update({"move_completed": True, "move_completed_path": settings.TORRENT_PATH})

        if settings.TORRENT_PAUSED:
            options.update({"add_paused": True})

        # TODO: Figure out a nice way to do this. Will currently only work if there is only one file in the download.
        # file_priorities (list of int): The priority for files in torrent, range is [0..7] however
        # only [0, 1, 4, 7] are normally used and correspond to [Skip, Low, Normal, High]
        priority_map = {-1: [1], 0: [4], 1: [7]}

        # result.priority =  -1 = low, 0 = normal, 1 = high
        options.update({"file_priorities": priority_map[result.priority]})
        # options.update({'file_priorities': priority_map[result.priority] * num_files})

        if result.ratio:
            options.update({"stop_at_ratio": True})
            options.update({"stop_ratio": float(result.ratio)})
            # options.update({'remove_at_ratio': True})
            # options.update({'remove_ratio': float(result.ratio)})

        return options



from urllib.parse import urljoin

import validators
from markdown2 import _slugify as slugify

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("MagnetDL")

        self.minseed = 0
        self.minleech = 0

        self.url = "http://www.magnetdl.com"
        self.urls = {"rss": urljoin(self.url, "download/tv/age/desc/")}

        self.custom_url = None
        self.headers.update({"Accept": "application/html"})

        self.cache = tvcache.TVCache(self)

    def search(self, search_strings, age=0, ep_obj=None):
        results = []

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))
                    search = slugify(search_string)
                    search_url = urljoin(self.url, "{}/{}/".format(search[0], search))
                else:
                    search_url = self.urls["rss"]

                if self.custom_url:
                    if not validators.url(self.custom_url):
                        logger.warning("Invalid custom url: {0}".format(self.custom_url))
                        return results
                    search_url = urljoin(self.custom_url, search_url.split(self.url)[1])

                data = self.get_url(search_url, returns="text")
                if not data:
                    logger.debug("URL did not return results/data, if the results are on the site maybe try a custom url, or a different one")
                    continue

                with BS4Parser(data, "html5lib") as html:
                    torrent_table = html.find("table", class_="download")
                    torrent_body = torrent_table.find("tbody") if torrent_table else []
                    # noinspection PyCallingNonCallable
                    torrent_rows = torrent_body("tr") if torrent_body else []

                    # Continue only if at least one Release is found
                    if not torrent_rows:
                        logger.debug("Data returned from provider does not contain any torrents")
                        continue

                    labels = [x.get_text(strip=True) for x in torrent_table.find("thead").find("tr")("th")]

                    # Skip column headers
                    for result in torrent_rows[0:-1:2]:
                        try:
                            if len(result("td")) < len(labels):
                                continue

                            title = result.find("td", class_="n").find("a")["title"]
                            magnet = result.find("td", class_="m").find("a")["href"]
                            seeders = try_int(result.find("td", class_="s").get_text(strip=True))
                            leechers = try_int(result.find("td", class_="l").get_text(strip=True))
                            size = convert_size(result("td")[labels.index("Size")].get_text(strip=True) or "") or -1

                            if not all([title, magnet]):
                                continue

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the"
                                        " minimum seeders or leechers: {0} (S:{1} L:{2})".format(title, seeders, leechers)
                                    )
                                continue

                            item = {"title": title, "link": magnet + self._custom_trackers, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)
                        except Exception as e:
                            continue

                # For each search mode sort all the items by seeders if available
                items.sort(key=lambda d: d.get("seeders", 0), reverse=True)
                results += items

            return results



class RTContext(object):
    def __init__(self):
        print("here")

    def get_conn(self):
        raise NotImplementedError

    def get_available_rpc_methods(self):
        raise NotImplementedError



import sickchill
from sickchill import logger, settings
from sickchill.oldbeard.trakt_api import TraktAPI, traktAuthException, traktException, traktServerBusy


class Notifier(object):
    """
    A "notifier" for trakt.tv which keeps track of what has and hasn't been added to your library.
    """

    def notify_snatch(self, ep_name):
        pass

    def notify_download(self, ep_name):
        pass

    def notify_subtitle_download(self, ep_name, lang):
        pass

    def notify_git_update(self, new_version):
        pass

    def notify_login(self, ipaddress=""):
        pass

    @staticmethod
    def update_library(ep_obj):
        """
        Sends a request to trakt indicating that the given episode is part of our library.

        ep_obj: The TVEpisode object to add to trakt
        """

        trakt_api = TraktAPI(settings.SSL_VERIFY, settings.TRAKT_TIMEOUT)

        if settings.USE_TRAKT:
            try:
                # URL parameters
                data = {
                    "shows": [
                        {
                            "title": ep_obj.show.name,
                            "year": ep_obj.show.startyear,
                            "ids": {ep_obj.idxr.slug: ep_obj.show.indexerid},
                        }
                    ]
                }

                if settings.TRAKT_SYNC_WATCHLIST and settings.TRAKT_REMOVE_SERIESLIST:
                    trakt_api.traktRequest("sync/watchlist/remove", data, method="POST")

                # Add Season and Episode + Related Episodes
                data["shows"][0]["seasons"] = [{"number": ep_obj.season, "episodes": []}]

                for relEp_Obj in [ep_obj] + ep_obj.relatedEps:
                    data["shows"][0]["seasons"][0]["episodes"].append({"number": relEp_Obj.episode})

                if settings.TRAKT_SYNC_WATCHLIST and settings.TRAKT_REMOVE_WATCHLIST:
                    trakt_api.traktRequest("sync/watchlist/remove", data, method="POST")

                # update library
                trakt_api.traktRequest("sync/collection", data, method="POST")

            except (traktException, traktAuthException, traktServerBusy) as e:
                logger.warning("Could not connect to Trakt service: {0}".format(str(e)))

    @staticmethod
    def update_watchlist(show_obj=None, s=None, e=None, data_show=None, data_episode=None, update="add"):
        """
        Sends a request to trakt indicating that the given episode is part of our library.

        show_obj: The TVShow object to add to trakt
        s: season number
        e: episode number
        data_show: structured object of shows trakt type
        data_episode: structured object of episodes trakt type
        update: type o action add or remove
        """

        trakt_api = TraktAPI(settings.SSL_VERIFY, settings.TRAKT_TIMEOUT)

        if settings.USE_TRAKT:

            data = {}
            try:
                # URL parameters
                if show_obj is not None:
                    data = {
                        "shows": [
                            {
                                "title": show_obj.name,
                                "year": show_obj.startyear,
                                "ids": {show_obj.idxr.slug: show_obj.indexerid},
                            }
                        ]
                    }
                elif data_show is not None:
                    data.update(data_show)
                else:
                    logger.warning("there's a coding problem contact developer. It's needed to be provided at lest one of the two: data_show or show_obj")
                    return False

                if data_episode is not None:
                    data["shows"][0].update(data_episode)

                elif s is not None:
                    # trakt URL parameters
                    season = {
                        "season": [
                            {
                                "number": s,
                            }
                        ]
                    }

                    if e is not None:
                        # trakt URL parameters
                        episode = {"episodes": [{"number": e}]}

                        season["season"][0].update(episode)

                    data["shows"][0].update(season)

                trakt_url = "sync/watchlist"
                if update == "remove":
                    trakt_url += "/remove"

                trakt_api.traktRequest(trakt_url, data, method="POST")

            except (traktException, traktAuthException, traktServerBusy) as e:
                logger.warning("Could not connect to Trakt service: {0}".format(str(e)))
                return False

        return True

    @staticmethod
    def trakt_show_data_generate(data):

        showList = []
        # TODO: is indexer and indexerid swapped here or in traktChecker:591? !Important
        for indexer, indexerid, title, year in data:
            show = {"title": title, "year": year, "ids": {sickchill.indexer.slug(indexer): indexerid}}
            showList.append(show)

        post_data = {"shows": showList}

        return post_data

    @staticmethod
    def trakt_episode_data_generate(data):

        # Find how many unique season we have
        uniqueSeasons = []
        for season, episode in data:
            if season not in uniqueSeasons:
                uniqueSeasons.append(season)

        # build the query
        seasonsList = []
        for searchedSeason in uniqueSeasons:
            episodesList = []
            for season, episode in data:
                if season == searchedSeason:
                    episodesList.append({"number": episode})
            seasonsList.append({"number": searchedSeason, "episodes": episodesList})

        post_data = {"seasons": seasonsList}

        return post_data

    @staticmethod
    def test_notify(username, blacklist_name=None):
        """
        Sends a test notification to trakt with the given authentication info and returns a boolean
        representing success.

        api: The api string to use
        username: The username to use
        blacklist_name: slug of trakt list used to hide not interested show

        Returns: True if the request succeeded, False otherwise
        """
        try:
            trakt_api = TraktAPI(settings.SSL_VERIFY, settings.TRAKT_TIMEOUT)
            trakt_api.validateAccount()
            if blacklist_name and blacklist_name is not None:
                trakt_lists = trakt_api.traktRequest("users/" + username + "/lists")
                found = False
                for trakt_list in trakt_lists:
                    if trakt_list["ids"]["slug"] == blacklist_name:
                        return "Test notice sent successfully to Trakt"
                if not found:
                    return "Trakt blacklist doesn't exists"
            else:
                return "Test notice sent successfully to Trakt"
        except (traktException, traktAuthException, traktServerBusy) as e:
            logger.warning("Could not connect to Trakt service: {0}".format(str(e)))
            return "Test notice failed to Trakt: {0}".format(str(e))



import json
import time

import certifi
import requests

from sickchill import logger, settings

from .exceptions import traktException


class TraktAPI:
    def __init__(self, ssl_verify=True, timeout=30):
        self.session = requests.Session()
        self.verify = certifi.where() if ssl_verify else False
        self.timeout = timeout if timeout else None
        self.auth_url = settings.TRAKT_OAUTH_URL
        self.api_url = settings.TRAKT_API_URL
        self.headers = {"Content-Type": "application/json", "trakt-api-version": "2", "trakt-api-key": settings.TRAKT_API_KEY}

    def traktToken(self, trakt_pin=None, refresh=False, count=0):

        if count > 3:
            settings.TRAKT_ACCESS_TOKEN = ""
            return False
        elif count > 0:
            time.sleep(2)

        data = {"client_id": settings.TRAKT_API_KEY, "client_secret": settings.TRAKT_API_SECRET, "redirect_uri": "urn:ietf:wg:oauth:2.0:oob"}

        if refresh:
            data["grant_type"] = "refresh_token"
            data["refresh_token"] = settings.TRAKT_REFRESH_TOKEN
        else:
            data["grant_type"] = "authorization_code"
            if trakt_pin:
                data["code"] = trakt_pin

        headers = {"Content-Type": "application/json"}

        resp = self.traktRequest("oauth/token", data=data, headers=headers, url=self.auth_url, method="POST", count=count)

        if "access_token" in resp:
            settings.TRAKT_ACCESS_TOKEN = resp["access_token"]
            if "refresh_token" in resp:
                settings.TRAKT_REFRESH_TOKEN = resp["refresh_token"]
            return True
        return False

    def validateAccount(self):

        resp = self.traktRequest("users/settings")

        if "account" in resp:
            return True
        return False

    def traktRequest(self, path, data=None, headers=None, url=None, method="GET", count=0):
        if url is None:
            url = self.api_url

        count = count + 1

        if headers is None:
            headers = self.headers

        if settings.TRAKT_ACCESS_TOKEN == "" and count >= 2:
            logger.warning("You must get a Trakt TOKEN. Check your Trakt settings")
            return {}

        if settings.TRAKT_ACCESS_TOKEN != "":
            headers["Authorization"] = "Bearer " + settings.TRAKT_ACCESS_TOKEN

        try:
            resp = self.session.request(method, url + path, headers=headers, timeout=self.timeout, data=json.dumps(data) if data else [], verify=self.verify)

            # check for http errors and raise if any are present
            resp.raise_for_status()

            # convert response to json
            resp = resp.json()
        except requests.exceptions.RequestException as e:
            code = getattr(e.response, "status_code", None)
            if not code:
                if "timed out" in e:
                    logger.warning("Timeout connecting to Trakt. Try to increase timeout value in Trakt settings")
                # This is pretty much a fatal error if there is no status_code
                # It means there basically was no response at all
                else:
                    logger.debug("Could not connect to Trakt. Error: {0}".format(e))
            elif code == 502:
                # Retry the request, cloudflare had a proxying issue
                logger.debug("Retrying trakt api request: %s" % path)
                return self.traktRequest(path, data, headers, url, method)
            elif code == 401:
                if self.traktToken(refresh=True, count=count):
                    return self.traktRequest(path, data, headers, url, method)
                else:
                    logger.warning("Unauthorized. Please check your Trakt settings")
            elif code in (500, 501, 503, 504, 520, 521, 522):
                # http://docs.trakt.apiary.io/#introduction/status-codes
                logger.debug("Trakt may have some issues and it's unavailable. Try again later please")
            elif code == 404:
                logger.debug("Trakt error (404) the resource does not exist: %s" % url + path)
            else:
                logger.exception("Could not connect to Trakt. Code error: {0}".format(code))
            return {}

        # check and confirm trakt call did not fail
        if isinstance(resp, dict) and resp.get("status", False) == "failure":
            if "message" in resp:
                raise traktException(resp["message"])
            if "error" in resp:
                raise traktException(resp["error"])
            else:
                raise traktException("Unknown Error")

        return resp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["току що", "съвсем скоро"],
    ["преди %s секунди", "след %s секунди"],
    ["преди 1 минута", "след 1 минута"],
    ["преди %s минути", "след %s минути"],
    ["преди 1 час", "след 1 час"],
    ["преди %s часа", "след %s часа"],
    ["преди 1 ден", "след 1 ден"],
    ["преди %s дни", "след %s дни"],
    ["преди 1 седмица", "след 1 седмица"],
    ["преди %s седмици", "след %s седмици"],
    ["преди 1 месец", "след 1 месец"],
    ["преди %s месеца", "след %s месеца"],
    ["преди 1 година", "след 1 година"],
    ["преди %s години", "след %s години"]
]
        


import datetime

from sickchill.show.History import History

from . import db
from .common import FAILED, Quality, SNATCHED, SUBTITLED


def _logHistoryItem(action, showid, season, episode, quality, resource, provider, version=-1):
    """
    Insert a history item in DB

    :param action: action taken (snatch, download, etc)
    :param showid: showid this entry is about
    :param season: show season
    :param episode: show episode
    :param quality: media quality
    :param resource: resource used
    :param provider: provider used
    :param version: tracked version of file (defaults to -1)
    """
    logDate = datetime.datetime.today().strftime(History.date_format)

    main_db_con = db.DBConnection()
    main_db_con.action(
        "INSERT INTO history (action, date, showid, season, episode, quality, resource, provider, version) VALUES (?,?,?,?,?,?,?,?,?)",
        [action, logDate, showid, season, episode, quality, resource, provider, version],
    )


def logSnatch(searchResult):
    """
    Log history of snatch

    :param searchResult: search result object
    """
    for curEpObj in searchResult.episodes:

        showid = int(curEpObj.show.indexerid)
        season = int(curEpObj.season)
        episode = int(curEpObj.episode)
        quality = searchResult.quality
        version = searchResult.version

        providerClass = searchResult.provider
        if providerClass is not None:
            provider = providerClass.name
        else:
            provider = "unknown"

        action = Quality.compositeStatus(SNATCHED, searchResult.quality)

        resource = searchResult.name

        _logHistoryItem(action, showid, season, episode, quality, resource, provider, version)


def logDownload(episode, filename, new_ep_quality, release_group=None, version=-1):
    """
    Log history of download

    :param episode: episode of show
    :param filename: file on disk where the download is
    :param new_ep_quality: Quality of download
    :param release_group: Release group
    :param version: Version of file (defaults to -1)
    """
    showid = int(episode.show.indexerid)
    season = int(episode.season)
    epNum = int(episode.episode)

    quality = new_ep_quality

    # store the release group as the provider if possible
    if release_group:
        provider = release_group
    else:
        provider = -1

    action = episode.status

    _logHistoryItem(action, showid, season, epNum, quality, filename, provider, version)


def logSubtitle(showid, season, episode, status, subtitleResult):
    """
    Log download of subtitle

    :param showid: Showid of download
    :param season: Show season
    :param episode: Show episode
    :param status: Status of download
    :param subtitleResult: Result object
    """
    resource = subtitleResult.language.opensubtitles
    provider = subtitleResult.provider_name

    status, quality = Quality.splitCompositeStatus(status)
    action = Quality.compositeStatus(SUBTITLED, quality)

    _logHistoryItem(action, showid, season, episode, quality, resource, provider)


def logFailed(epObj, release, provider=None):
    """
    Log a failed download

    :param epObj: Episode object
    :param release: Release group
    :param provider: Provider used for snatch
    """
    showid = int(epObj.show.indexerid)
    season = int(epObj.season)
    epNum = int(epObj.episode)
    status, quality = Quality.splitCompositeStatus(epObj.status)
    action = Quality.compositeStatus(FAILED, quality)

    _logHistoryItem(action, showid, season, epNum, quality, release, provider)



# coding=utf-8
import codecs
import glob
import json

from git import Repo
import os

import shutil
import re

from subprocess import call

# Note that you have to specify path to script

dest = './jsrepo'

# Clean dest
try:
    shutil.rmtree(dest)
except:
    pass

if not os.access(dest, os.F_OK):
    os.mkdir(dest)

Repo.clone_from('https://github.com/hustcc/timeago.js.git', dest)

for filename in glob.glob(os.path.join(dest, 'locales/*.js')):
    with open(filename) as f:
        content = f.readlines()
        # We're doing just the basic syntax and we're not trying to understand locale.js which is the index file
        if len(content) > 18 or 'locales.js' in filename:
            print('Seems that %s is a bit too complex for this parser' % filename)
            continue
        print('Generating %s\'s JSON Object...' % filename)
        call(["node", "convert_local_to_json.js", filename])

override = False

for filename in glob.glob('./tmp/*.json'):
    wrong = False
    pure_filename = filename.split('/')[-1:][0][:-5]
    if not override:
        for already_existing in glob.glob('./*.py'):
            if already_existing[2:-3] == pure_filename:
                wrong = True
                break
        if wrong:
            continue
    with open(filename) as g:
        json_value = json.loads(g.read())

        generated_placeholder = []

        for x in json_value['statements']:
            generated_placeholder.append(x[0])
            generated_placeholder.append(x[1])

        txt = \
"""#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"],
    ["%s", "%s"]
]
        """ % tuple(generated_placeholder)
        with codecs.open(pure_filename+'.py', 'w+', "utf-8") as h:
            h.write(txt)
            h.close()


# Clean
shutil.rmtree('./tmp')
shutil.rmtree(dest)




import re
from urllib.parse import quote_plus

from bs4 import BeautifulSoup
from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("HDSpace")

        self.username = None
        self.password = None
        self.minseed = 0
        self.minleech = 0

        self.cache = tvcache.TVCache(self, min_time=10)  # only poll HDSpace every 10 minutes max

        self.urls = {
            "base_url": "https://hd-space.org/",
            "login": "https://hd-space.org/index.php?page=login",
            "search": "https://hd-space.org/index.php?page=torrents&search=%s&active=1&options=0",
            "rss": "https://hd-space.org/rss_torrents.php?feed=dl",
        }

        self.categories = [15, 21, 22, 24, 25, 40]  # HDTV/DOC 1080/720, bluray, remux
        self.urls["search"] += "&category="
        for cat in self.categories:
            self.urls["search"] += str(cat) + "%%3B"
            self.urls["rss"] += "&cat[]=" + str(cat)
        self.urls["search"] = self.urls["search"][:-4]  # remove extra %%3B

        self.url = self.urls["base_url"]

    def _check_auth(self):

        if not self.username or not self.password:
            logger.warning("Invalid username or password. Check your settings")

        return True

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        if "pass" in dict_from_cookiejar(self.session.cookies):
            return True

        login_params = {"uid": self.username, "pwd": self.password}

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if re.search("Password Incorrect", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    search_url = self.urls["search"] % (quote_plus(search_string.replace(".", " ")),)
                else:
                    search_url = self.urls["search"] % ""

                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                data = self.get_url(search_url, returns="text")
                if not data or "please try later" in data:
                    logger.debug("No data returned from provider")
                    continue

                # Search result page contains some invalid html that prevents html parser from returning all data.
                # We cut everything before the table that contains the data we are interested in thus eliminating
                # the invalid html portions
                try:
                    data = data.split('<div id="information"></div>')[1]
                    index = data.index("<table")
                except ValueError:
                    logger.exception("Could not find main torrent table")
                    continue
                except IndexError:
                    logger.debug("Could not parse data from provider")
                    continue

                html = BeautifulSoup(data[index:], "html5lib")
                if not html:
                    logger.debug("No html data parsed from provider")
                    continue

                torrents = html("tr")
                if not torrents:
                    continue

                # Skip column headers
                for result in torrents[1:]:
                    if len(result.contents) < 10:
                        # skip extraneous rows at the end
                        continue

                    try:
                        dl_href = result.find("a", attrs={"href": re.compile(r"download.php.*")})["href"]
                        title = re.search("f=(.*).torrent", dl_href).group(1).replace("+", ".")
                        download_url = self.urls["base_url"] + dl_href
                        seeders = int(result.find("span", class_="seedy").find("a").text)
                        leechers = int(result.find("span", class_="leechy").find("a").text)
                        torrent_size = re.match(r".*?([0-9]+,?\.?[0-9]* [KkMmGg]+[Bb]+).*", str(result), re.DOTALL).group(1)
                        size = convert_size(torrent_size) or -1

                        if not all([title, download_url]):
                            continue

                        # Filter unseeded torrent
                        if seeders < self.minseed or leechers < self.minleech:
                            if mode != "RSS":
                                logger.debug(
                                    "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                        title, seeders, leechers
                                    )
                                )
                            continue

                        item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                        if mode != "RSS":
                            logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                        items.append(item)

                    except (AttributeError, TypeError, KeyError, ValueError):
                        continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



class SickChillException(Exception):
    """
    Generic SickChill Exception - should never be thrown, only sub-classed
    """


class AuthException(SickChillException):
    """
    Your authentication information are incorrect
    """


class CantRefreshShowException(SickChillException):
    """
    The show can't be refreshed right now
    """


class CantRemoveShowException(SickChillException):
    """
    The show can't removed right now
    """


class CantUpdateShowException(SickChillException):
    """
    The show can't be updated right now
    """


class EpisodeDeletedException(SickChillException):
    """
    This episode has been deleted
    """


class EpisodeNotFoundException(SickChillException):
    """
    The episode wasn't found on the Indexer
    """


class EpisodePostProcessingFailedException(SickChillException):
    """
    The episode post-processing failed
    """


class FailedPostProcessingFailedException(SickChillException):
    """
    The failed post-processing failed
    """


class MultipleEpisodesInDatabaseException(SickChillException):
    """
    Multiple episodes were found in the database! The database must be fixed first
    """


class MultipleShowsInDatabaseException(SickChillException):
    """
    Multiple shows were found in the database! The database must be fixed first
    """


class MultipleShowObjectsException(SickChillException):
    """
    Multiple objects for the same show were found! Something is very wrong
    """


class NoNFOException(SickChillException):
    """
    No NFO was found
    """


class ShowDirectoryNotFoundException(SickChillException):
    """
    The show directory was not found
    """


class ShowNotFoundException(SickChillException):
    """
    The show wasn't found on the Indexer
    """


class UpdaterException(SickChillException):
    """
    The updater encountered an error
    """



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
from calendar import timegm

try:
    from rfc822 import parsedate
except ImportError:
    from email.utils import parsedate


class TwitterModel(object):

    """ Base class from which all twitter models will inherit. """

    def __init__(self, **kwargs):
        self.param_defaults = {}

    def __str__(self):
        """ Returns a string representation of TwitterModel. By default
        this is the same as AsJsonString(). """
        return self.AsJsonString()

    def __eq__(self, other):
        return other and self.AsDict() == other.AsDict()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        if hasattr(self, 'id'):
            return hash(self.id)
        else:
            raise TypeError('unhashable type: {} (no id attribute)'
                            .format(type(self)))

    def AsJsonString(self, ensure_ascii=True):
        """ Returns the TwitterModel as a JSON string based on key/value
        pairs returned from the AsDict() method. """
        return json.dumps(self.AsDict(), ensure_ascii=ensure_ascii, sort_keys=True)

    def AsDict(self):
        """ Create a dictionary representation of the object. Please see inline
        comments on construction when dictionaries contain TwitterModels. """
        data = {}

        for (key, value) in self.param_defaults.items():

            # If the value is a list, we need to create a list to hold the
            # dicts created by an object supporting the AsDict() method,
            # i.e., if it inherits from TwitterModel. If the item in the list
            # doesn't support the AsDict() method, then we assign the value
            # directly. An example being a list of Media objects contained
            # within a Status object.
            if isinstance(getattr(self, key, None), (list, tuple, set)):
                data[key] = list()
                for subobj in getattr(self, key, None):
                    if getattr(subobj, 'AsDict', None):
                        data[key].append(subobj.AsDict())
                    else:
                        data[key].append(subobj)

            # Not a list, *but still a subclass of TwitterModel* and
            # and we can assign the data[key] directly with the AsDict()
            # method of the object. An example being a Status object contained
            # within a User object.
            elif getattr(getattr(self, key, None), 'AsDict', None):
                data[key] = getattr(self, key).AsDict()

            # If the value doesn't have an AsDict() method, i.e., it's not
            # something that subclasses TwitterModel, then we can use direct
            # assigment.
            elif getattr(self, key, None):
                data[key] = getattr(self, key, None)
        return data

    @classmethod
    def NewFromJsonDict(cls, data, **kwargs):
        """ Create a new instance based on a JSON dict. Any kwargs should be
        supplied by the inherited, calling class.

        Args:
            data: A JSON dict, as converted from the JSON in the twitter API.

        """

        json_data = data.copy()
        if kwargs:
            for key, val in kwargs.items():
                json_data[key] = val

        c = cls(**json_data)
        c._json = data
        return c


class Media(TwitterModel):

    """A class representing the Media component of a tweet. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            'display_url': None,
            'expanded_url': None,
            'ext_alt_text': None,
            'id': None,
            'media_url': None,
            'media_url_https': None,
            'sizes': None,
            'type': None,
            'url': None,
            'video_info': None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "Media(ID={media_id}, Type={media_type}, DisplayURL='{url}')".format(
            media_id=self.id,
            media_type=self.type,
            url=self.display_url)


class List(TwitterModel):

    """A class representing the List structure used by the twitter API. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            'description': None,
            'following': None,
            'full_name': None,
            'id': None,
            'member_count': None,
            'mode': None,
            'name': None,
            'slug': None,
            'subscriber_count': None,
            'uri': None,
            'user': None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

        if 'user' in kwargs:
            self.user = User.NewFromJsonDict(kwargs.get('user'))

    def __repr__(self):
        return "List(ID={list_id}, FullName={full_name!r}, Slug={slug}, User={user})".format(
            list_id=self.id,
            full_name=self.full_name,
            slug=self.slug,
            user=self.user.screen_name)


class Category(TwitterModel):

    """A class representing the suggested user category structure. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            'name': None,
            'size': None,
            'slug': None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "Category(Name={name!r}, Slug={slug}, Size={size})".format(
            name=self.name,
            slug=self.slug,
            size=self.size)


class DirectMessage(TwitterModel):

    """A class representing a Direct Message. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            'created_at': None,
            'id': None,
            'recipient_id': None,
            'sender_id': None,
            'text': None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        if self.text and len(self.text) > 140:
            text = "{text}[...]".format(text=self.text[:140])
        else:
            text = self.text
        return "DirectMessage(ID={dm_id}, Sender={sender}, Created={time}, Text='{text!r}')".format(
            dm_id=self.id,
            sender=self.sender_id,
            time=self.created_at,
            text=text)


class Trend(TwitterModel):

    """ A class representing a trending topic. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            'events': None,
            'name': None,
            'promoted_content': None,
            'query': None,
            'timestamp': None,
            'url': None,
            'tweet_volume': None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "Trend(Name={0!r}, Time={1}, URL={2})".format(
            self.name,
            self.timestamp,
            self.url)

    @property
    def volume(self):
        return self.tweet_volume


class Hashtag(TwitterModel):

    """ A class representing a twitter hashtag. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            'text': None
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "Hashtag(Text={text!r})".format(
            text=self.text)


class Url(TwitterModel):

    """ A class representing an URL contained in a tweet. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            'expanded_url': None,
            'url': None}

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "URL(URL={url}, ExpandedURL={eurl})".format(
            url=self.url,
            eurl=self.expanded_url)


class UserStatus(TwitterModel):

    """ A class representing the UserStatus structure. This is an abbreviated
    form of the twitter.User object. """

    _connections = {'following': False,
                    'followed_by': False,
                    'following_received': False,
                    'following_requested': False,
                    'blocking': False,
                    'muting': False}

    def __init__(self, **kwargs):
        self.param_defaults = {
            'blocking': False,
            'followed_by': False,
            'following': False,
            'following_received': False,
            'following_requested': False,
            'id': None,
            'id_str': None,
            'muting': False,
            'name': None,
            'screen_name': None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

        if 'connections' in kwargs:
            for param in self._connections:
                if param in kwargs['connections']:
                    setattr(self, param, True)

    @property
    def connections(self):
        return {'following': self.following,
                'followed_by': self.followed_by,
                'following_received': self.following_received,
                'following_requested': self.following_requested,
                'blocking': self.blocking,
                'muting': self.muting}

    def __repr__(self):
        connections = [param for param in self.connections if getattr(self, param)]
        return "UserStatus(ID={uid}, ScreenName={sn}, Connections=[{conn}])".format(
            uid=self.id,
            sn=self.screen_name,
            conn=", ".join(connections))


class User(TwitterModel):

    """A class representing the User structure. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            'contributors_enabled': None,
            'created_at': None,
            'default_profile': None,
            'default_profile_image': None,
            'description': None,
            'email': None,
            'favourites_count': None,
            'followers_count': None,
            'following': None,
            'friends_count': None,
            'geo_enabled': None,
            'id': None,
            'id_str': None,
            'lang': None,
            'listed_count': None,
            'location': None,
            'name': None,
            'notifications': None,
            'profile_background_color': None,
            'profile_background_image_url': None,
            'profile_background_image_url_https': None,
            'profile_background_tile': None,
            'profile_banner_url': None,
            'profile_image_url': None,
            'profile_image_url_https': None,
            'profile_link_color': None,
            'profile_sidebar_border_color': None,
            'profile_sidebar_fill_color': None,
            'profile_text_color': None,
            'profile_use_background_image': None,
            'protected': None,
            'screen_name': None,
            'status': None,
            'statuses_count': None,
            'time_zone': None,
            'url': None,
            'utc_offset': None,
            'verified': None,
            'withheld_in_countries': None,
            'withheld_scope': None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "User(ID={uid}, ScreenName={sn})".format(
            uid=self.id,
            sn=self.screen_name)

    @classmethod
    def NewFromJsonDict(cls, data, **kwargs):
        from twitter import Status
        if data.get('status', None):
            status = Status.NewFromJsonDict(data.get('status'))
            return super(cls, cls).NewFromJsonDict(data=data, status=status)
        else:
            return super(cls, cls).NewFromJsonDict(data=data)


class Status(TwitterModel):
    """A class representing the Status structure used by the twitter API.
    """

    def __init__(self, **kwargs):
        self.param_defaults = {
            'contributors': None,
            'coordinates': None,
            'created_at': None,
            'current_user_retweet': None,
            'favorite_count': None,
            'favorited': None,
            'full_text': None,
            'geo': None,
            'hashtags': None,
            'id': None,
            'id_str': None,
            'in_reply_to_screen_name': None,
            'in_reply_to_status_id': None,
            'in_reply_to_user_id': None,
            'lang': None,
            'location': None,
            'media': None,
            'place': None,
            'possibly_sensitive': None,
            'quoted_status': None,
            'quoted_status_id': None,
            'quoted_status_id_str': None,
            'retweet_count': None,
            'retweeted': None,
            'retweeted_status': None,
            'scopes': None,
            'source': None,
            'text': None,
            'truncated': None,
            'urls': None,
            'user': None,
            'user_mentions': None,
            'withheld_copyright': None,
            'withheld_in_countries': None,
            'withheld_scope': None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

        if kwargs.get('full_text', None):
            self.tweet_mode = 'extended'
        else:
            self.tweet_mode = 'compatibility'

    @property
    def created_at_in_seconds(self):
        """ Get the time this status message was posted, in seconds since
        the epoch (1 Jan 1970).

        Returns:
            int: The time this status message was posted, in seconds since
            the epoch.
        """
        return timegm(parsedate(self.created_at))

    def __repr__(self):
        """ A string representation of this twitter.Status instance.
        The return value is the ID of status, username and datetime.

        Returns:
            string: A string representation of this twitter.Status instance with
            the ID of status, username and datetime.
        """
        if self.tweet_mode == 'extended':
            text = self.full_text
        else:
            text = self.text
        if self.user:
            return "Status(ID={0}, ScreenName={1}, Created={2}, Text={3!r})".format(
                self.id,
                self.user.screen_name,
                self.created_at,
                text)
        else:
            return u"Status(ID={0}, Created={1}, Text={2!r})".format(
                self.id,
                self.created_at,
                text)

    @classmethod
    def NewFromJsonDict(cls, data, **kwargs):
        """ Create a new instance based on a JSON dict.

        Args:
            data: A JSON dict, as converted from the JSON in the twitter API

        Returns:
            A twitter.Status instance
        """
        current_user_retweet = None
        hashtags = None
        media = None
        quoted_status = None
        retweeted_status = None
        urls = None
        user = None
        user_mentions = None

        # for loading extended tweets from the streaming API.
        if 'extended_tweet' in data:
            for k, v in data['extended_tweet'].items():
                data[k] = v

        if 'user' in data:
            user = User.NewFromJsonDict(data['user'])
        if 'retweeted_status' in data:
            retweeted_status = Status.NewFromJsonDict(data['retweeted_status'])
        if 'current_user_retweet' in data:
            current_user_retweet = data['current_user_retweet']['id']
        if 'quoted_status' in data:
            quoted_status = Status.NewFromJsonDict(data.get('quoted_status'))

        if 'entities' in data:
            if 'urls' in data['entities']:
                urls = [Url.NewFromJsonDict(u) for u in data['entities']['urls']]
            if 'user_mentions' in data['entities']:
                user_mentions = [User.NewFromJsonDict(u) for u in data['entities']['user_mentions']]
            if 'hashtags' in data['entities']:
                hashtags = [Hashtag.NewFromJsonDict(h) for h in data['entities']['hashtags']]
            if 'media' in data['entities']:
                media = [Media.NewFromJsonDict(m) for m in data['entities']['media']]

        # the new extended entities
        if 'extended_entities' in data:
            if 'media' in data['extended_entities']:
                media = [Media.NewFromJsonDict(m) for m in data['extended_entities']['media']]

        return super(cls, cls).NewFromJsonDict(data=data,
                                               current_user_retweet=current_user_retweet,
                                               hashtags=hashtags,
                                               media=media,
                                               quoted_status=quoted_status,
                                               retweeted_status=retweeted_status,
                                               urls=urls,
                                               user=user,
                                               user_mentions=user_mentions)




import sys

# Syntax sugar.
_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)

from io import open as io_open

if is_py2:
    basestring = basestring
    str = unicode
    open = io_open
elif is_py3:
    basestring = (str, bytes)
    str = str
    open = open



import math
import socket
import time
from datetime import datetime

import jsonrpclib

from sickchill import logger, settings
from sickchill.helper.common import episode_num
from sickchill.helper.exceptions import AuthException
from sickchill.oldbeard import classes, scene_exceptions, tvcache
from sickchill.oldbeard.common import cpu_presets
from sickchill.oldbeard.helpers import sanitizeSceneName
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("BTN")

        self.supports_absolute_numbering = True

        self.api_key = None

        self.cache = BTNCache(self, min_time=15)  # Only poll BTN every 15 minutes max

        self.urls = {
            "base_url": "https://api.broadcasthe.net",
            "website": "http://broadcasthe.net/",
        }

        self.url = self.urls["website"]

    def _check_auth(self):
        if not self.api_key:
            logger.warning("Invalid api key. Check your settings")

        return True

    def _check_auth_from_data(self, parsed_json):

        if parsed_json is None:
            return self._check_auth()

        if "api-error" in parsed_json:
            logger.debug("Incorrect authentication credentials: {0}".format(parsed_json["api-error"]))
            raise AuthException("Your authentication credentials for " + self.name + " are incorrect, check your config.")

        return True

    def search(self, search_params, age=0, ep_obj=None):  # pylint:disable=too-many-locals

        self._check_auth()

        results = []
        params = {}
        apikey = self.api_key

        # age in seconds
        if age:
            params["age"] = "<=" + str(int(age))

        if search_params:
            params.update(search_params)
            logger.debug("Search string: {0}".format(search_params))

        parsed_json = self._api_call(apikey, params)
        if not parsed_json:
            logger.debug("No data returned from provider")
            return results

        if self._check_auth_from_data(parsed_json):

            if "torrents" in parsed_json:
                found_torrents = parsed_json["torrents"]
            else:
                found_torrents = {}

            # We got something, we know the API sends max 1000 results at a time.
            # See if there are more than 1000 results for our query, if not we
            # keep requesting until we've got everything.
            # max 150 requests per hour so limit at that. Scan every 15 minutes. 60 / 15 = 4.
            max_pages = 150
            results_per_page = 1000

            if "results" in parsed_json and int(parsed_json["results"]) >= results_per_page:
                pages_needed = int(math.ceil(int(parsed_json["results"]) / results_per_page))
                if pages_needed > max_pages:
                    pages_needed = max_pages

                # +1 because range(1,4) = 1, 2, 3
                for page in range(1, pages_needed + 1):
                    parsed_json = self._api_call(apikey, params, results_per_page, page * results_per_page)
                    # Note that this these are individual requests and might time out individually. This would result in 'gaps'
                    # in the results. There is no way to fix this though.
                    if "torrents" in parsed_json:
                        found_torrents.update(parsed_json["torrents"])

            for _, torrent_info in found_torrents.items():
                (title, url) = self._get_title_and_url(torrent_info)

                if title and url:
                    logger.debug("Found result: {0} ".format(title))
                    results.append(torrent_info)

        # FIXME SORT RESULTS
        return results

    def _api_call(self, apikey, params=None, results_per_page=1000, offset=0):

        server = jsonrpclib.Server(self.urls["base_url"])
        parsed_json = {}

        try:
            parsed_json = server.getTorrents(apikey, params or {}, int(results_per_page), int(offset))
            time.sleep(cpu_presets[settings.CPU_PRESET])

        except jsonrpclib.jsonrpc.ProtocolError as error:
            if error == (-32001, "Invalid API Key"):
                logger.warning("The API key you provided was rejected because it is invalid. Check your provider configuration.")
            elif error == (-32002, "Call Limit Exceeded"):
                logger.warning("You have exceeded the limit of 150 calls per hour, per API key which is unique to your user account")
            else:
                logger.exception("JSON-RPC protocol error while accessing provider. Error: {0} ".format(repr(error)))
            parsed_json = {"api-error": str(error)}
            return parsed_json

        except socket.timeout:
            logger.warning("Timeout while accessing provider")

        except socket.error as error:
            # Note that sometimes timeouts are thrown as socket errors
            logger.warning("Socket error while accessing provider. Error: {0} ".format(error[1]))

        except Exception as error:
            errorstring = str(error)
            if errorstring.startswith("<") and errorstring.endswith(">"):
                errorstring = errorstring[1:-1]
            logger.warning("Unknown error while accessing provider. Error: {0} ".format(errorstring))

        return parsed_json

    def _get_title_and_url(self, parsed_json):

        # The BTN API gives a lot of information in response,
        # however SickChill is built mostly around Scene or
        # release names, which is why we are using them here.

        if "ReleaseName" in parsed_json and parsed_json["ReleaseName"]:
            title = parsed_json["ReleaseName"]
            append = ""
            if "Resolution" in parsed_json and parsed_json["Resolution"].lower() not in title.lower():
                append += parsed_json["Resolution"]
            if "Source" in parsed_json and parsed_json["Source"].lower() not in title.lower():
                append += parsed_json["Source"]
            if "Codec" in parsed_json and parsed_json["Codec"].lower() not in title.lower():
                append += parsed_json["Codec"]

            if append:
                title += " [" + append + "]"

        else:
            # If we don't have a release name we need to get creative
            title = ""
            if "Series" in parsed_json:
                title += parsed_json["Series"]
            if "GroupName" in parsed_json:
                title += "." + parsed_json["GroupName"] if title else parsed_json["GroupName"]
            if "Resolution" in parsed_json:
                title += "." + parsed_json["Resolution"] if title else parsed_json["Resolution"]
            if "Source" in parsed_json:
                title += "." + parsed_json["Source"] if title else parsed_json["Source"]
            if "Codec" in parsed_json:
                title += "." + parsed_json["Codec"] if title else parsed_json["Codec"]
            if title:
                title = title.replace(" ", ".")

        url = None
        if "DownloadURL" in parsed_json:
            url = parsed_json["DownloadURL"]
            if url:
                # unescaped / is valid in JSON, but it can be escaped
                url = url.replace("\\/", "/")

        return title, url

    def get_season_search_strings(self, ep_obj):
        search_params = []
        current_params = {"category": "Season"}

        # Search for entire seasons: no need to do special things for air by date or sports shows
        if ep_obj.show.air_by_date or ep_obj.show.sports:
            # Search for the year of the air by date show
            current_params["name"] = str(ep_obj.airdate).split("-")[0]
        else:
            # BTN uses the same format for both Anime and TV
            current_params["name"] = "Season " + str(ep_obj.scene_season)

        # search
        if ep_obj.show.indexer == 1:
            current_params["tvdb"] = ep_obj.show.indexerid
            search_params.append(current_params)
        else:
            name_exceptions = list(set(scene_exceptions.get_scene_exceptions(ep_obj.show.indexerid) + [ep_obj.show.name]))
            for name in name_exceptions:
                # Search by name if we don't have tvdb id
                current_params["series"] = sanitizeSceneName(name)
                search_params.append(current_params)

        return search_params

    def get_episode_search_strings(self, ep_obj, add_string=""):

        if not ep_obj:
            return [{}]

        to_return = []
        search_params = {"category": "Episode"}

        # episode
        if ep_obj.show.air_by_date or ep_obj.show.sports:
            date_str = str(ep_obj.airdate)

            # BTN uses dots in dates, we just search for the date since that
            # combined with the series identifier should result in just one episode
            search_params["name"] = date_str.replace("-", ".")
        else:
            # BTN uses the same format for both Anime and TV
            # Do a general name search for the episode, formatted like SXXEYY
            search_params["name"] = "{ep}".format(ep=episode_num(ep_obj.scene_season, ep_obj.scene_episode))

        # search
        if ep_obj.show.indexer == 1:
            search_params["tvdb"] = ep_obj.show.indexerid
            to_return.append(search_params)
        else:
            # add new query string for every exception
            name_exceptions = list(set(scene_exceptions.get_scene_exceptions(ep_obj.show.indexerid) + [ep_obj.show.name]))
            for cur_exception in name_exceptions:
                search_params["series"] = sanitizeSceneName(cur_exception)
                to_return.append(search_params)

        return to_return

    def _do_general_search(self, search_string):
        # 'search' looks as broad is it can find. Can contain episode overview and title for example,
        # use with caution!
        return self.search({"search": search_string})

    def find_propers(self, search_date=None):
        results = []

        search_terms = ["%.proper.%", "%.repack.%"]

        for term in search_terms:
            for item in self.search({"release": term}, age=4 * 24 * 60 * 60):
                if item["Time"]:
                    try:
                        result_date = datetime.fromtimestamp(float(item["Time"]))
                    except TypeError:
                        result_date = None

                    if result_date and (not search_date or result_date > search_date):
                        title, url = self._get_title_and_url(item)
                        if title and url:
                            results.append(classes.Proper(title, url, result_date, self.show))

        return results


class BTNCache(tvcache.TVCache):
    def _get_rss_data(self):
        # Get the torrents uploaded since last check.
        seconds_since_last_update = math.ceil(time.time() - time.mktime(self.last_update.timetuple()))

        # default to 15 minutes
        seconds_minTime = self.min_time * 60
        if seconds_since_last_update < seconds_minTime:
            seconds_since_last_update = seconds_minTime

        # Set maximum to 24 hours (24 * 60 * 60 = 86400 seconds) of "RSS" data search, older things will need to be done through backlog
        if seconds_since_last_update > 86400:
            logger.debug("The last known successful update was more than 24 hours ago, only trying to fetch the last 24 hours!")
            seconds_since_last_update = 86400

        self.search_params = None  # BTN cache does not use search params
        return {"entries": self.provider.search(search_params=self.search_params, age=seconds_since_last_update)}



import logging
import os.path
import sys
import unittest
from collections import namedtuple

from configobj import ConfigObj

from sickchill import settings
from sickchill.oldbeard import config, scheduler


class ConfigTestBasic(unittest.TestCase):
    """
    Test basic methods in oldbeard.config
    """

    def test_check_section(self):
        """
        Test check_section
        """
        CFG = ConfigObj("config.ini", encoding="UTF-8", indent_type="  ")
        assert not config.check_section(CFG, "General")
        assert config.check_section(CFG, "General")

    def test_checkbox_to_value(self):
        """
        Test checkbox_to_value
        """
        assert config.checkbox_to_value(1)
        assert config.checkbox_to_value(["option", "True"])
        assert config.checkbox_to_value("0", "yes", "no") == "no"

    def test_clean_host(self):
        """
        Test clean_host
        """
        assert config.clean_host("http://127.0.0.1:8080") == "127.0.0.1:8080"
        assert config.clean_host("https://mail.google.com/mail") == "mail.google.com"
        assert config.clean_host("http://localhost:8081/home/displayShow?show=80379#season-10") == "localhost:8081"
        assert config.clean_host("http://testme.co.uk", 9000) == "testme.co.uk:9000"  # default port
        assert config.clean_host("www.google.com/search") == "www.google.com"
        assert config.clean_host("") == ""  # empty host

    def test_clean_hosts(self):
        """
        Test clean_hosts
        """
        dirty_hosts = (
            "http://127.0.0.1:8080,https://mail.google.com/mail," "http://localhost:8081/home/displayShow?show=80379#season-10," "www.google.com/search,"
        )
        clean_result = "127.0.0.1:8080,mail.google.com:5050,localhost:8081,www.google.com:5050"
        assert config.clean_hosts(dirty_hosts, "5050") == clean_result

    def test_clean_url(self):
        """
        Test cleaning of urls
        """
        log = logging.getLogger(__name__)
        test = namedtuple("test", "expected_result dirty clean")

        url_tests = [
            test(True, "https://subdomain.domain.tld/endpoint", "https://subdomain.domain.tld/endpoint"),  # does not add a final /
            test(True, "http://www.example.com/folder/", "http://www.example.com/folder/"),  # does not remove the final /
            test(True, "google.com/xml.rpc", "http://google.com/xml.rpc"),  # add scheme if missing
            test(True, "google.com", "http://google.com/"),  # add scheme if missing and final / if its just the domain
            test(True, "scgi:///home/user/.config/path/socket", "scgi:///home/user/.config/path/socket"),  # scgi identified as scheme
            test(True, None, ""),  # None URL returns empty string
            test(False, "https://subdomain.domain.tld/endpoint", "http://subdomain.domain.tld/endpoint"),  # does not change schemes from https to http
            test(False, "http://subdomain.domain.tld/endpoint", "https://subdomain.domain.tld/endpoint"),  # ...or vice versa
            test(False, "google.com/xml.rpc", "google.com/xml.rpc"),  # scheme is always added
            test(False, "google.com", "https://google.com/"),  # does not default to https
            test(False, "http://www.example.com/folder/", "http://www.example.com/folder"),  # does not strip final /
            test(False, "scgi:///home/user/.config/path/socket", "scgi:///home/user/.config/path/socket/"),  # does not add a final /
            test(AttributeError, 1, 1),  # None URL returns empty string
        ]

        for test_url in url_tests:
            if issubclass(type(Exception), type(test_url.expected_result)):
                with self.assertRaises(test_url.expected_result):
                    assert config.clean_url(test_url.dirty) == test_url.clean
            elif test_url.expected_result is True:
                assert config.clean_url(test_url.dirty) == test_url.clean
            elif test_url.expected_result is False:
                assert config.clean_url(test_url.dirty) != test_url.clean
            else:
                log.error("Test not defined for %s", test_url)

    def test_min_max(self):
        """
        Test min_max
        """

        assert config.min_max("100", default=50, low=50, high=200) == 100
        assert config.min_max("25", default=50, low=50, high=200) == 50
        assert config.min_max("250", default=50, low=50, high=200) == 200

    def test_check_setting_int(self):
        """
        Test check_setting_int
        """
        # setup
        CFG = ConfigObj("config.ini", encoding="UTF-8", indent_type="  ")
        config.check_section(CFG, "General")
        CFG["General"]["indexer_timeout"] = 60
        CFG["General"]["use_icacls"] = "True"
        CFG["General"]["use_nzbs"] = "False"
        CFG["General"]["status_default"] = None
        # normal
        assert config.check_setting_int(CFG, "General", "indexer_timeout", 30) == 60
        assert CFG["General"]["indexer_timeout"] == 60
        # force min/max
        assert config.check_setting_int(CFG, "General", "indexer_timeout", 150, 100, 200) == 150
        assert CFG["General"]["indexer_timeout"] == 150
        assert config.check_setting_int(CFG, "General", "indexer_timeout", 250, 200, 300, False) == 200
        assert CFG["General"]["indexer_timeout"] == 200
        assert config.check_setting_int(CFG, "General", "indexer_timeout", 90, 50, 100) == 90
        assert CFG["General"]["indexer_timeout"] == 90
        assert config.check_setting_int(CFG, "General", "indexer_timeout", 20, 10, 30, False) == 30
        assert CFG["General"]["indexer_timeout"] == 30
        # true/false => int
        assert config.check_setting_int(CFG, "General", "use_icacls", 1) == 1
        assert CFG["General"]["use_icacls"] == "True"
        assert config.check_setting_int(CFG, "General", "use_nzbs", 0) == 0
        assert CFG["General"]["use_nzbs"] == "False"
        # None value type + silent off
        assert config.check_setting_int(CFG, "General", "status_default", 5, silent=False) == 5
        assert CFG["General"]["status_default"] == 5
        # unmatched section
        assert config.check_setting_int(CFG, "Subtitles", "subtitles_finder_frequency", 1) == 1
        assert CFG["Subtitles"]["subtitles_finder_frequency"] == 1
        # wrong def_val/min/max type
        assert config.check_setting_int(CFG, "General", "indexer_timeout", "ba", "min", "max") == 30
        assert CFG["General"]["indexer_timeout"] == 30

    def test_check_setting_float(self):
        """
        Test check_setting_float
        """
        # setup
        CFG = ConfigObj("config.ini", encoding="UTF-8", indent_type="  ")
        config.check_section(CFG, "General")
        CFG["General"]["fanart_background_opacity"] = 0.5
        CFG["General"]["log_size"] = None
        # normal
        assert config.check_setting_float(CFG, "General", "fanart_background_opacity", 0.4) == 0.5
        assert CFG["General"]["fanart_background_opacity"] == 0.5
        # force min/max
        assert config.check_setting_float(CFG, "General", "fanart_background_opacity", 0.7, 0.6, 1.0) == 0.7
        assert CFG["General"]["fanart_background_opacity"] == 0.7
        assert config.check_setting_float(CFG, "General", "fanart_background_opacity", 0.7, 0.8, 1.0, False) == 0.8
        assert CFG["General"]["fanart_background_opacity"] == 0.8
        assert config.check_setting_float(CFG, "General", "fanart_background_opacity", 0.3, 0.1, 0.4) == 0.3
        assert CFG["General"]["fanart_background_opacity"] == 0.3
        assert config.check_setting_float(CFG, "General", "fanart_background_opacity", 0.1, 0.1, 0.2, False) == 0.2
        assert CFG["General"]["fanart_background_opacity"] == 0.2
        # None value type + silent off
        assert config.check_setting_float(CFG, "General", "log_size", 10.0, silent=False) == 10.0
        assert CFG["General"]["log_size"] == 10.0
        # unmatched section
        assert config.check_setting_float(CFG, "Kodi", "log_size", 2.5) == 2.5
        assert CFG["Kodi"]["log_size"] == 2.5
        # wrong def_val/min/max type
        assert config.check_setting_float(CFG, "General", "fanart_background_opacity", "ba", "min", "max") == 0.2
        assert CFG["General"]["fanart_background_opacity"] == 0.2

    def test_check_setting_str(self):
        """
        Test check_setting_str
        """
        # setup
        CFG = ConfigObj("config.ini", encoding="UTF-8", indent_type="  ")
        config.check_section(CFG, "General")
        CFG["General"]["process_method"] = "copy"
        CFG["General"]["git_password"] = "SFa342FHb_"  # noqa: S105
        CFG["General"]["extra_scripts"] = None
        # normal
        assert config.check_setting_str(CFG, "General", "process_method", "move") == "copy"
        assert config.check_setting_str(CFG, "General", "git_password", "", silent=False, censor_log=True) == "SFa342FHb_"
        # None value type
        assert config.check_setting_str(CFG, "General", "extra_scripts", "") == ""
        # unmatched section
        assert config.check_setting_str(CFG, "Subtitles", "subtitles_languages", "eng") == "eng"
        # wrong def_val type
        assert config.check_setting_str(CFG, "General", "process_method", ["fail"]) == "copy"

    def test_check_setting_bool(self):
        """
        Test check_setting_bool
        """
        # setup
        CFG = ConfigObj("config.ini", encoding="UTF-8", indent_type="  ")
        config.check_section(CFG, "General")
        CFG["General"]["debug"] = True
        CFG["General"]["season_folders_default"] = False
        CFG["General"]["dbdebug"] = None
        # normal
        assert config.check_setting_bool(CFG, "General", "debug")
        assert not config.check_setting_bool(CFG, "General", "season_folders_default", def_val=True)
        # None value type
        assert not config.check_setting_bool(CFG, "General", "dbdebug", False)
        # unmatched item
        assert config.check_setting_bool(CFG, "General", "git_reset", def_val=True)
        # unmatched section
        assert not config.check_setting_bool(CFG, "Subtitles", "use_subtitles", def_val=False)
        # wrong def_val type, silent = off
        assert config.check_setting_bool(CFG, "General", "debug", def_val=["fail"], silent=False)


class ConfigTestChanges(unittest.TestCase):
    """
    Test change methods in oldbeard.config
    """

    def test_change_https_cert(self):
        """
        Test change_https_cert
        """
        settings.HTTPS_CERT = "server.crt"  # Initialize
        assert config.change_https_cert("")
        assert config.change_https_cert("server.crt")
        assert not config.change_https_cert("/:/server.crt")  # INVALID
        settings.HTTPS_CERT = ""

    def test_change_https_key(self):
        """
        Test change_https_key
        """
        settings.HTTPS_KEY = "server.key"  # Initialize
        assert config.change_https_key("")
        assert config.change_https_key("server.key")
        assert not config.change_https_key("/:/server.key")  # INVALID
        settings.HTTPS_KEY = ""

    def test_change_sickchill_background(self):
        """
        Test change_sickchill_background
        """
        settings.SICKCHILL_BACKGROUND_PATH = ""  # Initialize
        assert config.change_sickchill_background(__file__)
        assert not config.change_sickchill_background("not_real.jpg")
        assert config.change_sickchill_background("")

    def test_change_custom_css(self):
        """
        Test change_custom_css
        """
        settings.CUSTOM_CSS_PATH = ""  # Initialize
        assert not config.change_custom_css(__file__)  # not a css file
        assert not config.change_custom_css("not_real.jpg")  # doesn't exist
        assert not config.change_custom_css("sickchill_tests")  # isn't a file
        css_file = os.path.join(os.path.dirname(__file__), "custom.css")
        with open(css_file, "w") as f:
            f.write("table.main {\n    width: 100%;\n}")
        assert config.change_custom_css(css_file)  # real
        os.remove(css_file)
        assert config.change_custom_css("")  # empty

    def test_change_nzb_dir(self):
        """
        Test change_nzb_dir
        """
        settings.NZB_DIR = ""
        assert config.change_nzb_dir("cache")
        assert not config.change_nzb_dir("/:/NZB_Downloads")  # INVALID
        assert config.change_nzb_dir("")

    def test_change_torrent_dir(self):
        """
        Test change_torrent_dir
        """
        settings.TORRENT_DIR = ""
        assert config.change_torrent_dir("cache")
        assert not config.change_torrent_dir("/:/Downloads")  # INVALID
        assert config.change_torrent_dir("")

    def test_change_tv_download_dir(self):
        """
        Test change_tv_download_dir
        """
        settings.TV_DOWNLOAD_DIR = ""
        assert config.change_tv_download_dir("cache")
        assert not config.change_tv_download_dir("/:/Downloads/Completed")  # INVALID

        assert config.change_tv_download_dir("")
        assert settings.TV_DOWNLOAD_DIR == ""

    def test_change_unpack_dir(self):
        """
        Test change_unpack_dir
        """
        settings.UNPACK_DIR = ""
        assert config.change_unpack_dir("cache")
        assert not config.change_unpack_dir("/:/Extract")  # INVALID
        assert config.change_unpack_dir("")

    def test_change_auto_pp_freq(self):
        """
        Test change_postprocessor_frequency
        """
        settings.autoPostProcessorScheduler = scheduler.Scheduler(lambda: None)  # dummy

        config.change_postprocessor_frequency(0)
        assert settings.AUTOPOSTPROCESSOR_FREQUENCY == settings.MIN_AUTOPOSTPROCESSOR_FREQUENCY
        config.change_postprocessor_frequency("s")
        assert settings.AUTOPOSTPROCESSOR_FREQUENCY == settings.DEFAULT_AUTOPOSTPROCESSOR_FREQUENCY
        config.change_postprocessor_frequency(60)
        assert settings.AUTOPOSTPROCESSOR_FREQUENCY == 60

    def test_change_daily_search_freq(self):
        """
        Test change_daily_search_frequency
        """
        settings.dailySearchScheduler = scheduler.Scheduler(lambda: None)  # dummy

        config.change_daily_search_frequency(0)
        assert settings.DAILYSEARCH_FREQUENCY == settings.MIN_DAILYSEARCH_FREQUENCY
        config.change_daily_search_frequency("s")
        assert settings.DAILYSEARCH_FREQUENCY == settings.DEFAULT_DAILYSEARCH_FREQUENCY
        config.change_daily_search_frequency(60)
        assert settings.DAILYSEARCH_FREQUENCY == 60

    def test_change_backlog_freq(self):
        """
        Test change_backlog_frequency
        """
        settings.backlogSearchScheduler = scheduler.Scheduler(lambda: None)  # dummy
        settings.DAILYSEARCH_FREQUENCY = settings.DEFAULT_DAILYSEARCH_FREQUENCY  # needed

        config.change_backlog_frequency(0)
        assert settings.BACKLOG_FREQUENCY == settings.MIN_BACKLOG_FREQUENCY
        config.change_backlog_frequency("s")
        assert settings.BACKLOG_FREQUENCY == settings.MIN_BACKLOG_FREQUENCY
        config.change_backlog_frequency(1440)
        assert settings.BACKLOG_FREQUENCY == 1440

    def test_change_update_freq(self):
        """
        Test change_update_frequency
        """
        settings.versionCheckScheduler = scheduler.Scheduler(lambda: None)  # dummy

        config.change_update_frequency(0)
        assert settings.UPDATE_FREQUENCY == settings.MIN_UPDATE_FREQUENCY
        config.change_update_frequency("s")
        assert settings.UPDATE_FREQUENCY == settings.DEFAULT_UPDATE_FREQUENCY
        config.change_update_frequency(60)
        assert settings.UPDATE_FREQUENCY == 60

    def test_change_show_update_hour(self):
        """
        Test change_showupdate_hour
        """
        settings.showUpdateScheduler = scheduler.Scheduler(lambda: None)  # dummy

        config.change_showupdate_hour(-2)
        assert settings.SHOWUPDATE_HOUR == 0
        config.change_showupdate_hour("s")
        assert settings.SHOWUPDATE_HOUR == settings.DEFAULT_SHOWUPDATE_HOUR
        config.change_showupdate_hour(60)
        assert settings.SHOWUPDATE_HOUR == 0
        config.change_showupdate_hour(12)
        assert settings.SHOWUPDATE_HOUR == 12

    def test_change_sub_finder_freq(self):
        """
        Test change_subtitle_finder_frequency
        """
        config.change_subtitle_finder_frequency("")
        assert settings.SUBTITLES_FINDER_FREQUENCY == 1
        config.change_subtitle_finder_frequency("s")
        assert settings.SUBTITLES_FINDER_FREQUENCY == 1
        config.change_subtitle_finder_frequency(8)
        assert settings.SUBTITLES_FINDER_FREQUENCY == 8

    def test_change_version_notify(self):
        """
        Test change_version_notify
        """

        class dummy_action(object):  # needed for *scheduler.action.forceRun()
            def __init__(self):
                self.amActive = False

        settings.versionCheckScheduler = scheduler.Scheduler(dummy_action())  # dummy
        settings.VERSION_NOTIFY = True

        config.change_version_notify(True)  # no change
        assert settings.VERSION_NOTIFY
        config.change_version_notify("stop")  # = defaults to False
        assert not settings.VERSION_NOTIFY
        assert not settings.versionCheckScheduler.enable
        config.change_version_notify("on")
        assert settings.VERSION_NOTIFY
        assert settings.versionCheckScheduler.enable

    def test_change_download_propers(self):
        """
        Test change_download_propers
        """
        settings.properFinderScheduler = scheduler.Scheduler(lambda: None)  # dummy
        settings.DOWNLOAD_PROPERS = True

        config.change_download_propers(True)  # no change
        assert settings.DOWNLOAD_PROPERS
        config.change_download_propers("stop")  # = defaults to False
        assert not settings.DOWNLOAD_PROPERS
        assert not settings.properFinderScheduler.enable
        config.change_download_propers("on")
        assert settings.DOWNLOAD_PROPERS
        assert settings.properFinderScheduler.enable

    def test_change_use_trakt(self):
        """
        Test change_use_trakt
        """
        settings.traktCheckerScheduler = scheduler.Scheduler(lambda: None)  # dummy
        settings.USE_TRAKT = True

        config.change_use_trakt(True)  # no change
        assert settings.USE_TRAKT
        config.change_use_trakt("stop")  # = defaults to False
        assert not settings.USE_TRAKT
        assert not settings.traktCheckerScheduler.enable
        config.change_use_trakt("on")
        assert settings.USE_TRAKT
        assert settings.traktCheckerScheduler.enable

    def test_change_use_subtitles(self):
        """
        Test change_use_subtitles
        """
        settings.subtitlesFinderScheduler = scheduler.Scheduler(lambda: None)  # dummy
        settings.USE_SUBTITLES = True

        config.change_use_subtitles(True)  # no change
        assert settings.USE_SUBTITLES
        config.change_use_subtitles("stop")  # = defaults to False
        assert not settings.USE_SUBTITLES
        assert not settings.subtitlesFinderScheduler.enable
        config.change_use_subtitles("on")
        assert settings.USE_SUBTITLES
        assert settings.subtitlesFinderScheduler.enable

    def test_change_process_auto(self):
        """
        Test change_process_automatically
        """
        settings.autoPostProcessorScheduler = scheduler.Scheduler(lambda: None)  # dummy
        settings.PROCESS_AUTOMATICALLY = True

        config.change_process_automatically(True)  # no change
        assert settings.PROCESS_AUTOMATICALLY
        config.change_process_automatically("stop")  # = defaults to False
        assert not settings.PROCESS_AUTOMATICALLY
        assert not settings.autoPostProcessorScheduler.enable
        config.change_process_automatically("on")
        assert settings.PROCESS_AUTOMATICALLY
        assert settings.autoPostProcessorScheduler.enable


class ConfigTestMigrator(unittest.TestCase):
    """
    Test the oldbeard.config.ConfigMigrator class
    """

    @unittest.expectedFailure  # Not fully implemented
    def test_config_migrator(self):
        """
        Test migrate_config
        """
        # TODO: Assert the 'too-advanced-config-version' error

        CFG = ConfigObj("config.ini", encoding="UTF-8", indent_type="  ")
        config.check_section(CFG, "General")
        CFG["General"]["config_version"] = 0
        settings.CONFIG_VERSION = 13
        settings.CONFIG_FILE = "config.ini"

        migrator = config.ConfigMigrator(CFG)
        migrator.migrate_config()


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger(__name__).setLevel(logging.DEBUG)

    SUITE = unittest.TestLoader().loadTestsFromTestCase(ConfigTestBasic)
    unittest.TextTestRunner(verbosity=2).run(SUITE)
    SUITE = unittest.TestLoader().loadTestsFromTestCase(ConfigTestChanges)
    unittest.TextTestRunner(verbosity=2).run(SUITE)
    SUITE = unittest.TestLoader().loadTestsFromTestCase(ConfigTestMigrator)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



import os
import sys
from argparse import ArgumentParser


class SickChillArgumentParser(ArgumentParser):
    def error(self, message):
        sys.stderr.write("error: {}\n".format(message))
        self.print_help()
        sys.exit(2)

    def __init__(self, data_dir):
        super().__init__()
        self.data_dir = data_dir

        self.description = """SickChill is an automatic tv library manager. It handles searching, sending to your download client, organizing, renaming,
        and adding images and metadata. It handles it all (with a little bit of magic) so you don't have to.
        (c) 2017 SickChill
        """

        self.add_argument("-q", "--quiet", action="store_true", help="disable logging to the console")
        self.add_argument("--nolaunch", action="store_true", help="suppress launching the web browser on startup")
        self.add_argument("-p", "--port", type=int, help="the port to listen on")
        self.add_argument(
            "--datadir",
            help="full path to a folder where the database, config, cache and log files should be stored. Default: {data_dir}"
            "{sep}".format(data_dir=self.data_dir, sep=os.sep),
        )
        self.add_argument(
            "--config",
            help="full file path to override the default configuration file. Default: {data_dir}{sep}config.ini".format(data_dir=self.data_dir, sep=os.sep),
        )
        self.add_argument("--pidfile", help="combined with --daemon creates a pid file (full path)")
        self.add_argument("--noresize", action="store_true", help="prevent resizing of show images even if PIL is installed")
        daemon_help = "run as daemon (includes options --quiet --nolaunch)"
        if sys.platform in ["win32", "darwin"]:
            daemon_help = "running as daemon is not supported on your platform. it is substituted with: --quiet --nolaunch"
        self.add_argument("-d", "--daemon", action="store_true", help=daemon_help)
        self.add_argument(
            "--force-update",
            action="store_true",
            help="download the latest stable version and force an " "update (use when you're unable to do so using " "the web ui)",
        )
